from djblets.testing.testcases import TestCase
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.resources.mixins.api_tokens import ResourceAPITokenMixin
from djblets.webapi.resources.root import RootResource
from djblets.webapi.models import BaseWebAPIToken


class PolicyTestResource(ResourceAPITokenMixin, WebAPIResource):
    policy_id = 'test'


class SomeObjectResource(ResourceAPITokenMixin, WebAPIResource):
    policy_id = 'someobject'


root_resource = RootResource([
    SomeObjectResource(),
])


class APIPolicyWebAPIToken(BaseWebAPIToken):
    @classmethod
    def get_root_resource(self):
        return root_resource


class APIPolicyTests(TestCase):
    """Tests API policy through WebAPITokens."""
    def setUp(self):
        super(APIPolicyTests, self).setUp()

        self.resource = PolicyTestResource()

    def test_default_policy(self):
        """Testing API policy enforcement with default policy"""
        self.assert_policy(
            {},
            allowed_methods=['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'DELETE'])

    def test_global_allow_all(self):
        """Testing API policy enforcement with *.allow=*"""
        self.assert_policy(
            {
                '*': {
                    'allow': ['*'],
                }
            },
            allowed_methods=['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'DELETE'])

    def test_global_block_all(self):
        """Testing API policy enforcement with *.block=*"""
        self.assert_policy(
            {
                '*': {
                    'block': ['*'],
                }
            },
            blocked_methods=['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'DELETE'])

    def test_global_block_all_and_resource_allow_all(self):
        """Testing API policy enforcement with *.block=* and
        <resource>.*.allow=*
        """
        self.assert_policy(
            {
                '*': {
                    'block': ['*'],
                },
                'test': {
                    '*': {
                        'allow': ['*'],
                    },
                }
            },
            allowed_methods=['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'DELETE'])

    def test_global_allow_all_and_resource_block_all(self):
        """Testing API policy enforcement with *.allow=* and
        <resource>.*.block=*
        """
        self.assert_policy(
            {
                '*': {
                    'allow': ['*'],
                },
                'test': {
                    '*': {
                        'block': ['*'],
                    },
                }
            },
            blocked_methods=['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'DELETE'])

    def test_global_block_all_and_resource_all_allow_methods(self):
        """Testing API policy enforcement with *.block=* and
        <resource>.*.allow=[methods]
        """
        self.assert_policy(
            {
                '*': {
                    'block': ['*'],
                },
                'test': {
                    '*': {
                        'allow': ['GET', 'PUT'],
                    },
                }
            },
            allowed_methods=['GET', 'PUT'],
            blocked_methods=['HEAD', 'POST', 'PATCH', 'DELETE'])

    def test_global_allow_all_and_resource_all_block_specific(self):
        """Testing API policy enforcement with *.allow=* and
        <resource>.*.block=[methods]
        """
        self.assert_policy(
            {
                '*': {
                    'allow': ['*'],
                },
                'test': {
                    '*': {
                        'block': ['GET', 'PUT'],
                    },
                }
            },
            allowed_methods=['HEAD', 'POST', 'PATCH', 'DELETE'],
            blocked_methods=['GET', 'PUT'])

    def test_resource_block_all_and_allow_methods(self):
        """Testing API policy enforcement with <resource>.*.block=* and
        <resource>.*.allow=[methods] for specific methods
        """
        self.assert_policy(
            {
                'test': {
                    '*': {
                        'block': ['*'],
                        'allow': ['GET', 'PUT'],
                    }
                }
            },
            allowed_methods=['GET', 'PUT'],
            blocked_methods=['HEAD', 'POST', 'PATCH', 'DELETE'])

    def test_resource_allow_all_and_block_methods(self):
        """Testing API policy enforcement with <resource>.*.allow=* and
        <resource>.*.block=[methods] for specific methods
        """
        self.assert_policy(
            {
                'test': {
                    '*': {
                        'allow': ['*'],
                        'block': ['GET', 'PUT'],
                    },
                }
            },
            allowed_methods=['HEAD', 'POST', 'DELETE'],
            blocked_methods=['GET', 'PUT'])

    def test_id_allow_all(self):
        """Testing API policy enforcement with <resource>.<id>.allow=*"""
        self.assert_policy(
            {
                'test': {
                    '42': {
                        'allow': ['*'],
                    }
                }
            },
            resource_id=42,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_id_block_all(self):
        """Testing API policy enforcement with <resource>.<id>.block=*"""
        policy = {
            'test': {
                '42': {
                    'block': ['*'],
                }
            }
        }

        self.assert_policy(
            policy,
            resource_id=42,
            blocked_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

        self.assert_policy(
            policy,
            resource_id=100,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_resource_block_all_and_id_allow_all(self):
        """Testing API policy enforcement with <resource>.*.block=* and
        <resource>.<id>.allow=*
        """
        policy = {
            'test': {
                '*': {
                    'block': ['*'],
                },
                '42': {
                    'allow': ['*'],
                }
            }
        }

        self.assert_policy(
            policy,
            resource_id=42,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

        self.assert_policy(
            policy,
            resource_id=100,
            blocked_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_resource_allow_all_and_id_block_all(self):
        """Testing API policy enforcement with <resource>.<id>.allow=* and
        <resource>.<id>.block=*
        """
        policy = {
            'test': {
                '*': {
                    'allow': ['*'],
                },
                '42': {
                    'block': ['*'],
                }
            }
        }

        self.assert_policy(
            policy,
            resource_id=42,
            blocked_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

        self.assert_policy(
            policy,
            resource_id=100,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_global_block_all_and_id_allow_all(self):
        """Testing API policy enforcement with *.<id>.block=* and
        <resource>.<id>.allow=*
        """
        self.assert_policy(
            {
                '*': {
                    'block': ['*'],
                },
                'test': {
                    '42': {
                        'allow': ['*'],
                    }
                }
            },
            resource_id=42,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_global_allow_all_and_id_block_all(self):
        """Testing API policy enforcement with *.<id>.allow=* and
        <resource>.<id>.block=*"""
        policy = {
            '*': {
                'allow': ['*'],
            },
            'test': {
                '42': {
                    'block': ['*'],
                }
            }
        }

        self.assert_policy(
            policy,
            resource_id=42,
            blocked_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

        self.assert_policy(
            policy,
            resource_id=100,
            allowed_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def test_policy_methods_conflict(self):
        """Testing API policy enforcement with methods conflict"""
        self.assert_policy(
            {
                'test': {
                    '*': {
                        'allow': ['*'],
                        'block': ['*'],
                    },
                }
            },
            blocked_methods=['HEAD', 'GET', 'POST', 'PUT', 'DELETE'])

    def assert_policy(self, policy, allowed_methods=[], blocked_methods=[],
                      resource_id=None):
        if resource_id is not None:
            resource_id = str(resource_id)

        for method in allowed_methods:
            allowed = self.resource.is_resource_method_allowed(
                policy, method, resource_id)

            if not allowed:
                self.fail('Expected %s to be allowed, but was blocked'
                          % method)

        for method in blocked_methods:
            allowed = self.resource.is_resource_method_allowed(
                policy, method, resource_id)

            if allowed:
                self.fail('Expected %s to be blocked, but was allowed'
                          % method)


class APIPolicyValidationTests(TestCase):
    """Tests API policy validation."""
    def test_empty(self):
        """Testing BaseWebAPIToken.validate_policy with empty policy"""
        APIPolicyWebAPIToken.validate_policy({})

    def test_not_object(self):
        """Testing BaseWebAPIToken.validate_policy without JSON object"""
        self.assertRaisesValidationError(
            'The policy must be a JSON object.',
            APIPolicyWebAPIToken.validate_policy,
            [])

    #
    # Top-level 'resources' object
    #

    def test_no_resources_section(self):
        """Testing BaseWebAPIToken.validate_policy with non-empty policy and
        no resources section
        """
        self.assertRaisesValidationError(
            'The policy is missing a "resources" section.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'foo': {}
            })

    def test_resources_empty(self):
        """Testing BaseWebAPIToken.validate_policy with empty resources section
        """
        self.assertRaisesValidationError(
            'The policy\'s "resources" section must not be empty.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {}
            })

    def test_resources_invalid_format(self):
        """Testing BaseWebAPIToken.validate_policy with resources not an object
        """
        self.assertRaisesValidationError(
            'The policy\'s "resources" section must be a JSON object.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': []
            })

    #
    # '*' section
    #

    def test_global_valid(self):
        """Testing BaseWebAPIToken.validate_policy with valid '*' section"""
        APIPolicyWebAPIToken.validate_policy({
            'resources': {
                '*': {
                    'allow': ['*'],
                    'block': ['POST'],
                }
            }
        })

    def test_empty_global(self):
        """Testing BaseWebAPIToken.validate_policy with empty '*' section"""
        self.assertRaisesValidationError(
            'The "resources.*" section must have "allow" and/or "block" '
            'rules.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    '*': {}
                }
            })

    def test_global_not_object(self):
        """Testing BaseWebAPIToken.validate_policy with '*' section not a
        JSON object
        """
        self.assertRaisesValidationError(
            'The "resources.*" section must be a JSON object.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    '*': []
                }
            })

    def test_global_allow_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with *.allow not a list"""
        self.assertRaisesValidationError(
            'The "resources.*" section\'s "allow" rule must be a list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    '*': {
                        'allow': {}
                    }
                }
            })

    def test_global_block_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with *.block not a list"""
        self.assertRaisesValidationError(
            'The "resources.*" section\'s "block" rule must be a list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    '*': {
                        'block': {}
                    }
                }
            })

    #
    # resource-specific '*' section
    #

    def test_resource_global_valid(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.* valid"""
        APIPolicyWebAPIToken.validate_policy({
            'resources': {
                'someobject': {
                    '*': {
                        'allow': ['*'],
                        'block': ['POST'],
                    },
                }
            }
        })

    def test_resource_global_empty(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.* empty"""
        self.assertRaisesValidationError(
            'The "resources.someobject.*" section must have "allow" and/or '
            '"block" rules.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '*': {}
                    }
                }
            })

    def test_resource_global_invalid_policy_id(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.* with
        invalid policy ID
        """
        self.assertRaisesValidationError(
            '"foobar" is not a valid resource policy ID.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'foobar': {
                        '*': {
                            'allow': ['*'],
                        }
                    }
                }
            })

    def test_resource_global_not_object(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.* not an
        object
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.*" section must be a JSON object.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '*': []
                    }
                }
            })

    def test_resource_global_allow_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.*.allow not
        a list
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.*" section\'s "allow" rule must be a '
            'list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '*': {
                            'allow': {}
                        }
                    }
                }
            })

    def test_resource_global_block_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.*.block not
        a list
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.*" section\'s "block" rule must be a '
            'list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '*': {
                            'block': {}
                        }
                    }
                }
            })

    #
    # resource-specific ID section
    #

    def test_resource_id_valid(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id> valid
        """
        APIPolicyWebAPIToken.validate_policy({
            'resources': {
                'someobject': {
                    '42': {
                        'allow': ['*'],
                        'block': ['POST'],
                    },
                }
            }
        })

    def test_resource_id_empty(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id> empty
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.42" section must have "allow" and/or '
            '"block" rules.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '42': {}
                    }
                }
            })

    def test_resource_id_invalid_id_type(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id> with
        invalid ID type
        """
        self.assertRaisesValidationError(
            '42 must be a string in "resources.someobject"',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        42: {
                            'allow': ['*'],
                        }
                    }
                }
            })

    def test_resource_id_not_object(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id> not an
        object
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.42" section must be a JSON object.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '42': []
                    }
                }
            })

    def test_resource_id_allow_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id>.allow
        not a list
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.42" section\'s "allow" rule must '
            'be a list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '42': {
                            'allow': {}
                        }
                    }
                }
            })

    def test_resource_id_block_not_list(self):
        """Testing BaseWebAPIToken.validate_policy with <resource>.<id>.block
        not a list
        """
        self.assertRaisesValidationError(
            'The "resources.someobject.42" section\'s "block" rule must '
            'be a list.',
            APIPolicyWebAPIToken.validate_policy,
            {
                'resources': {
                    'someobject': {
                        '42': {
                            'block': {}
                        }
                    }
                }
            })



"""Utilities for rate-limiting login attempts."""

import logging
import re
import time

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured

from djblets.cache.backend import make_cache_key


logger = logging.getLogger(__name__)


#: Available rate limit categories.
RATE_LIMIT_LOGIN = 0
RATE_LIMIT_API_ANONYMOUS = 1
RATE_LIMIT_API_AUTHENTICATED = 2


#: The default rate limit for logins.
DEFAULT_LOGIN_LIMIT_RATE = '5/m'

#: The default rate limit for anonymous API requests.
DEFAULT_API_ANONYMOUS_LIMIT_RATE = '1000/h'

#: The default rate limit for authenticated API requests.
DEFAULT_API_AUTHENTICATED_LIMIT_RATE = '10000/h'


_RATE_LIMIT_DATA = {
    RATE_LIMIT_LOGIN: (
        'LOGIN_LIMIT_RATE',
        DEFAULT_LOGIN_LIMIT_RATE,
        'login-ratelimit'),
    RATE_LIMIT_API_ANONYMOUS: (
        'API_ANONYMOUS_LIMIT_RATE',
        DEFAULT_API_ANONYMOUS_LIMIT_RATE,
        'api-anonymous-ratelimit'),
    RATE_LIMIT_API_AUTHENTICATED: (
        'API_AUTHENTICATED_LIMIT_RATE',
        DEFAULT_API_AUTHENTICATED_LIMIT_RATE,
        'api-authenticated-ratelimit'),
}


def get_user_id_or_ip(request):
    """Return the user's ID or IP address from the given HTTP request.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        unicode:
        If the user is authenticated, the user ID will be returned.
        Otherwise, the IP address of the client is returned instead.
    """
    if hasattr(request, 'user') and request.user.is_authenticated:
        return str(request.user.pk)

    try:
        return request.META['HTTP_X_REAL_IP']
    except KeyError:
        try:
            return request.META['HTTP_X_FORWARDED_FOR'].split(',')[0].strip()
        except KeyError:
            return request.META['REMOTE_ADDR']


def _get_time_int():
    """Return the current time as an integer rounded to the nearest second.

    This is available as a convenience wrapper to help with unit testing.

    Version Added:
        2.3.4

    Returns:
        int:
        The current time rounded to the nearest second.
    """
    return int(time.time())


def _get_window(period, timestamp):
    """Return window period within the given time period.

    This helps determine the time left before the rate limit is over
    for the given user (see ``time_left`` variable in
    :py:func:`get_usage_count`).

    Args:
        period (int):
            The total time period in seconds from the rate.

        timestamp (int):
            The time in seconds used as "now" for the window calculation.

            Version Added:
                2.3.4

    Returns:
        int:
        The window period.
    """
    if period == 1:
        return timestamp

    return timestamp - (timestamp % period) + period


def is_ratelimited(request, increment=False, limit_type=RATE_LIMIT_LOGIN):
    """Check whether the user or IP address has exceeded the rate limit.

    The parameters are used to create a new key or fetch an existing key to
    save or update to the cache and to determine the amount of time period
    left.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        increment (bool, optional):
            Whether the number of login attempts should be incremented.

        limit_type (int, optional):
            The type of rate limit to check.

    Returns:
        bool:
        Whether the current user has exceeded the rate limit of login attempts.
    """
    usage = get_usage_count(request, increment, limit_type)
    return (usage is not None and
            usage['count'] > usage['limit'])


def get_usage_count(request, increment=False, limit_type=RATE_LIMIT_LOGIN):
    """Return rate limit status for a given user or IP address.

    This method performs validation checks on the input parameters
    and creates the cache key to keep track of the
    number of login attempts made by the user. It saves the new
    cache key and initial number of attempts or updates the
    existing cache key and number of attempts before returning
    the count, limit, and time_left.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        increment (bool, optional):
            Whether the number of login attempts should be incremented.

        limit_type (int, optional):
            The type of rate limit to check.

    Returns:
        dict:
        A dictionary with the following keys:

        ``count`` (:py:class:`int`):
            The number of login attempts made.

        ``limit`` (:py:class:`int`):
            The number of attempts allowed.

        ``time_left`` (:py:class:`int`):
            The time left before rate limit is over.
    """
    try:
        try:
            settings_key, default_value, cache_key_prefix = \
                _RATE_LIMIT_DATA[limit_type]
        except KeyError:
            raise ValueError('"limit_type" argument had unexpected value "%s"'
                             % limit_type)

        limit_str = getattr(settings, settings_key, default_value)

        if limit_str is None:
            # If the setting is explicitly None, don't do any rate limiting.
            return None

        rate_limit = Rate.parse(limit_str)
    except ValueError:
        raise ImproperlyConfigured('LOGIN_LIMIT_RATE setting could not '
                                   'be parsed.')

    limit = rate_limit.count
    period = rate_limit.seconds

    # Determine user ID or IP address from HTTP request.
    user_id_or_ip = get_user_id_or_ip(request)

    # Prepare cache key to add or update to cache and determine remaining time
    # period left.
    now = _get_time_int()
    window = _get_window(period, now)
    cache_key = make_cache_key('%s:%d/%d%s%s'
                               % (cache_key_prefix, limit, period,
                                  user_id_or_ip, window))
    time_left = window - now

    count = None

    if increment:
        try:
            count = cache.incr(cache_key)
        except ValueError:
            cache.add(cache_key, 1)

    if count is None:
        count = cache.get(cache_key, 0)

    if not increment:
        # Add one to the returned value, even if we aren't incrementing the
        # stored value. This makes it so that we're consistent in how many
        # tries per period regardless of whether we're incrementing now or
        # later.
        count += 1

    return {
        'count': count,
        'limit': limit,
        'time_left': time_left,
    }


class Rate(object):
    """A rate representing login attempt frequency.

    The main functionality of this class is found in the :py:meth:`parse`
    function. This class converts a rate into a Rate object, which
    contains the number of login attempts allowed within a time period based
    on a given rate string.
    """
    #: Dictionary contains keys that represent different time periods.
    #:
    #: The time periods are represented in seconds for days, hours,
    #: minutes, and seconds.

    PERIODS = {
        's': 1,
        'm': 60,
        'h': 60 * 60,
        'd': 24 * 60 * 60,
    }

    #: Regular expression that interprets the rate string.
    RATE_RE = re.compile(r'(\d+)/(\d*)([smhd])?')

    @classmethod
    def parse(cls, rate_str):
        """Return a Rate parsed from the given rate string.

        Converts the given rate string into a Rate object, which contains
        the number of login attempts allowed (count) and the time period
        alotted for these attempts (seconds).

        Args:
            rate (unicode):
                The number of attempts allowed within a period
                of time (can be seconds, minutes, hours, or days).

        Returns:
            Rate:
            A Rate object that returns the number of login attempts
            allowed (count), and the total time period for these attempts
            (seconds).
        """
        m = Rate.RATE_RE.match(rate_str)

        if m:
            count, multiplier, period = m.groups()
        else:
            raise ValueError('Could not parse given rate: %s.' % rate_str)

        seconds = Rate.PERIODS[period or 's']

        if multiplier:
            seconds *= int(multiplier)

        return cls(count=int(count), seconds=seconds)

    def __init__(self, count, seconds):
        """Initialize attributes for the Rate object.

        This initializes the number of failed login attempts allowed, and
        the time period for the login attempts in seconds based on the
        data returned from the parse() function.

        Args:
            count (int):
                The number of failed login attempts allowed.

            seconds (int):
                The time period for the login attempts in seconds.
        """
        self.count = count
        self.seconds = seconds

    def __eq__(self, other):
        """Return whether the two Rate instances are equal.

        Returns:
            bool:
            Return true if the count and seconds match.
        """
        return (isinstance(other, Rate) and
                self.count == other.count and
                self.seconds == other.seconds)



"""Utilities for working with gravatars."""

from hashlib import md5
from urllib.parse import urlencode

from django.conf import settings


def get_gravatar_url_for_email(email=None, size=None):
    """Return the Gravatar URL for an e-mail address.

    The returned URL will always use HTTPS.

    Note that callers adhering to the GDPR should check for a user's consent
    before displaying a Gravatar on their behalf. This is checked automatically
    if using :py:mod:`djblets.avatars`.

    Args:
        email (unicode):
            The e-mail address to get the Gravatar URL for.

        size (int, optional):
            An optional height and width of the image (in pixels).

    Returns:
        unicode:
        The URL for the Gravatar associated with the given e-mail address.
    """
    if email:
        email = email.strip().lower()

        if isinstance(email, str):
            email = email.encode('utf-8')

        email_hash = md5(email).hexdigest()
    else:
        email_hash = '00000000000000000000000000000000'

    url = 'https://secure.gravatar.com/avatar/%s' % email_hash
    params = []

    if not size and hasattr(settings, 'GRAVATAR_SIZE'):
        size = settings.GRAVATAR_SIZE

    if size:
        params.append(('s', size))

    if hasattr(settings, 'GRAVATAR_RATING'):
        params.append(('r', settings.GRAVATAR_RATING))

    if hasattr(settings, 'GRAVATAR_DEFAULT'):
        params.append(('d', settings.GRAVATAR_DEFAULT))

    if params:
        url = '%s?%s' % (url, urlencode(params))

    return url


def get_gravatar_url(user=None, size=None):
    """Return the Gravatar URL for a user.

    The returned URL will always use HTTPS.

    Note that callers adhering to the GDPR should check for a user's consent
    before displaying a Gravatar on their behalf. This is checked automatically
    if using :py:mod:`djblets.avatars`.

    Args:
        user (django.contrib.auth.models.User):
            The user whose Gravatar URL is to be retrieved.

        size (int, optional):
            An optional height and width of the image (in pixels).

    Returns:
        unicode:
        The URL for the user's Gravatar.

    Raises:
        ValueError:
            Raised if ``user`` is ``None``.
    """
    if user is None:
        raise ValueError('"user" cannot be None.')

    if user.is_anonymous or not user.email:
        return ''

    return get_gravatar_url_for_email(email=user.email, size=size)



"""Utility functions for working with template-related caches."""

from django.template import Engine, engines


def clear_template_tag_caches():
    """Clear the template tags caches.

    This allows changes to the list of available template tags to be reflected
    in any new templates by emptying all the caches and forcing the available
    list of tags to be rebuilt.
    """
    engines._engines.clear()
    Engine.get_default.cache_clear()


def clear_template_caches():
    """Clear the templates caches.

    This clears any caches for template parse trees and related state, forcing
    templates to be re-parsed and re-rendered.
    """
    template_loaders = []

    for engine in engines.all():
        template_loaders += engine.engine.template_loaders

    for template_loader in template_loaders:
        template_loader.reset()



"""Signals for being notified on registry operations."""

from django.dispatch import Signal


#: Emitted when a registry is populating.
#:
#: Args:
#:     registry (djblets.registries.registry.Registry):
#:         The registry being populated.
registry_populating = Signal()



"""Unit tests for djblets.siteconfig."""

import hmac

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.cache import cache

from djblets.siteconfig.django_settings import (apply_django_settings,
                                                cache_settings_map,
                                                mail_settings_map)
from djblets.siteconfig.models import SiteConfiguration
from djblets.siteconfig.signals import siteconfig_reloaded
from djblets.testing.testcases import TestCase


class SiteConfigTestCase(TestCase):
    """Base class for SiteConfiguration-related unit tests."""

    def setUp(self):
        super(SiteConfigTestCase, self).setUp()

        self.siteconfig = SiteConfiguration(site=Site.objects.get_current())
        self.siteconfig.save()

    def tearDown(self):
        super(SiteConfigTestCase, self).tearDown()

        self.siteconfig.delete()
        SiteConfiguration.objects.clear_cache()


class DjangoSettingsTests(SiteConfigTestCase):
    """Unit tests for django_settings.py."""

    def test_mail_auth_deserialize(self):
        """Testing Django mail siteconfig settings with deserializing mail
        server credentials
        """
        # This is bug 1476. We deserialized the e-mail settings to Unicode
        # strings automatically, but this broke mail sending on some setups.
        # The HMAC library is incompatible with Unicode strings in more recent
        # Python 2.6 versions. Now we deserialize as a string. This test
        # ensures that these settings never break again.

        username = 'myuser'
        password = 'mypass'

        self.assertEqual(type(username), str)
        self.assertEqual(type(password), str)

        self.siteconfig.set('mail_host_user', username)
        self.siteconfig.set('mail_host_password', password)
        apply_django_settings(self.siteconfig, mail_settings_map)

        self.assertEqual(settings.EMAIL_HOST_USER, username)
        self.assertEqual(settings.EMAIL_HOST_PASSWORD, password)
        self.assertEqual(type(settings.EMAIL_HOST_USER), str)
        self.assertEqual(type(settings.EMAIL_HOST_PASSWORD), str)

        # Simulate the failure point in HMAC
        settings.EMAIL_HOST_USER.translate(hmac.trans_5C)
        settings.EMAIL_HOST_PASSWORD.translate(hmac.trans_5C)

    def test_cache_backend_with_legacy_uri(self):
        """Testing Django cache backend siteconfig settings with migrating
        stored legacy cache backend URI to settings.CACHES
        """
        settings.CACHES = {
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
                'LOCATION': 'foo',
            },
            'staticfiles': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
                'LOCATION': 'staticfiles-cache',
            }
        }

        self.siteconfig.set('cache_backend', 'memcached://localhost:12345/')
        apply_django_settings(self.siteconfig, cache_settings_map)

        self.assertTrue('staticfiles' in settings.CACHES)
        self.assertTrue('default' in settings.CACHES)
        self.assertTrue('forwarded_backend' in settings.CACHES)

        self.assertEqual(
            settings.CACHES['default']['BACKEND'],
            'djblets.cache.forwarding_backend.ForwardingCacheBackend')
        self.assertEqual(settings.CACHES['default']['LOCATION'],
                         'forwarded_backend')

        self.assertEqual(
            settings.CACHES['forwarded_backend']['BACKEND'],
            'django.core.cache.backends.memcached.PyMemcacheCache')
        self.assertEqual(settings.CACHES['forwarded_backend']['LOCATION'],
                         'localhost:12345')

    def test_cache_backend_with_caches(self):
        """Testing Django cache backend siteconfig settings with stored CACHES
        dictionary overriding entry in settings.CACHES
        """
        settings.CACHES['staticfiles'] = {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'staticfiles-cache',
        }

        self.siteconfig.set('cache_backend', {
            'default': {
                'BACKEND':
                    'django.core.cache.backends.memcached.MemcachedCache',
                'LOCATION': 'localhost:12345',
            },
        })

        apply_django_settings(self.siteconfig, cache_settings_map)

        self.assertTrue('staticfiles' in settings.CACHES)
        self.assertTrue('default' in settings.CACHES)
        self.assertTrue('forwarded_backend' in settings.CACHES)

        self.assertEqual(
            settings.CACHES['default']['BACKEND'],
            'djblets.cache.forwarding_backend.ForwardingCacheBackend')
        self.assertEqual(settings.CACHES['default']['LOCATION'],
                         'forwarded_backend')

        self.assertEqual(
            settings.CACHES['forwarded_backend']['BACKEND'],
            'django.core.cache.backends.memcached.PyMemcacheCache')
        self.assertEqual(settings.CACHES['forwarded_backend']['LOCATION'],
                         'localhost:12345')

    def test_cache_backend_without_caches_in_settings_py(self):
        """Testing Django cache backend siteconfig settings with stored
        CACHES dictionary and no entry in settings.CACHES
        """
        settings.CACHES['staticfiles'] = {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'staticfiles-cache',
        }

        self.siteconfig.set('cache_backend', {
            'default': {
                'BACKEND':
                    'django.core.cache.backends.memcached.CacheClass',
                'LOCATION': 'localhost:12345',
            },
        })

        apply_django_settings(self.siteconfig, cache_settings_map)

        self.assertTrue('staticfiles' in settings.CACHES)
        self.assertTrue('default' in settings.CACHES)
        self.assertTrue('forwarded_backend' in settings.CACHES)

        self.assertEqual(
            settings.CACHES['default']['BACKEND'],
            'djblets.cache.forwarding_backend.ForwardingCacheBackend')
        self.assertEqual(settings.CACHES['default']['LOCATION'],
                         'forwarded_backend')

        self.assertEqual(
            settings.CACHES['forwarded_backend']['BACKEND'],
            'django.core.cache.backends.memcached.PyMemcacheCache')
        self.assertEqual(settings.CACHES['forwarded_backend']['LOCATION'],
                         'localhost:12345')

    def test_cache_backend_pymemcache(self):
        """Testing transition of old memcached backend to pymemcache"""
        settings.CACHES['staticfiles'] = {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'staticfiles-cache',
        }

        self.siteconfig.set('cache_backend', {
            'default': {
                'BACKEND':
                    'django.core.cache.backends.memcached.MemcachedCache',
                'LOCATION': 'localhost:12345',
            },
        })

        apply_django_settings(self.siteconfig, cache_settings_map)

        self.assertTrue('staticfiles' in settings.CACHES)
        self.assertTrue('default' in settings.CACHES)
        self.assertTrue('forwarded_backend' in settings.CACHES)

        self.assertEqual(
            settings.CACHES['default']['BACKEND'],
            'djblets.cache.forwarding_backend.ForwardingCacheBackend')
        self.assertEqual(settings.CACHES['default']['LOCATION'],
                         'forwarded_backend')

        self.assertEqual(
            settings.CACHES['forwarded_backend']['BACKEND'],
            'django.core.cache.backends.memcached.PyMemcacheCache')
        self.assertEqual(settings.CACHES['forwarded_backend']['LOCATION'],
                         'localhost:12345')


class SiteConfigurationTests(SiteConfigTestCase):
    """Unit tests for SiteConfiguration."""

    def test_get(self):
        """Testing SiteConfiguration.get"""
        self.siteconfig.set('valid_key_2', 'valid_parameter_2')

        self.assertEqual(self.siteconfig.get('valid_key_2'),
                         'valid_parameter_2')

    def test_get_with_missing_key_and_explicit_default(self):
        """Testing SiteConfiguration.get and passing a missing key with
        explicit default
        """
        self.assertEqual(self.siteconfig.get('invalid_key', default='default'),
                         'default')

    def test_get_with_missing_key_and_registered_default(self):
        """Testing SiteConfiguration.get and passing a missing key with a
        registered default
        """
        SiteConfiguration.add_global_default('valid_key_1', 'global_default_1')
        self.siteconfig.add_default('valid_key_1', 'valid_parameter_1')

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'valid_parameter_1')
        finally:
            self.siteconfig.remove_default('valid_key_1')

    def test_get_with_missing_key_and_registered_global_default(self):
        """Testing SiteConfiguration.get and passing a missing key with a
        registered global default
        """
        SiteConfiguration.add_global_default('valid_key_1', 'global_default_1')

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'global_default_1')
        finally:
            SiteConfiguration.remove_global_default('valid_key_1')

    def test_set(self):
        """Testing SiteConfiguration.set"""
        self.siteconfig.set('valid_key_2', 'valid_parameter_2')
        self.siteconfig.set('valid_key_3', 'valid_parameter_3')

        self.assertEqual(
            self.siteconfig.settings,
            {
                'valid_key_2': 'valid_parameter_2',
                'valid_key_3': 'valid_parameter_3',
            })

    def test_add_defaults(self):
        """Testing SiteConfiguration.add_defaults"""
        self.siteconfig.add_defaults({
            'valid_key_1': 'valid_parameter_1',
            'valid_key_2': 'valid_parameter_2',
            'valid_key_3': 'valid_parameter_3',
        })

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'valid_parameter_1')
            self.assertEqual(self.siteconfig.get('valid_key_2'),
                             'valid_parameter_2')
            self.assertEqual(self.siteconfig.get('valid_key_3'),
                             'valid_parameter_3')
        finally:
            self.siteconfig.remove_default('valid_key_1')
            self.siteconfig.remove_default('valid_key_2')
            self.siteconfig.remove_default('valid_key_3')

    def test_add_default(self):
        """Testing SiteConfiguration.add_default"""
        self.siteconfig.add_default('valid_key_1', 'valid_new_parameter_2')

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'valid_new_parameter_2')
        finally:
            self.siteconfig.remove_default('valid_key_1')

    def test_remove_default(self):
        """Testing SiteConfiguration.remove_default"""
        self.siteconfig.add_default('valid_key_1', 'valid_new_parameter_2')
        self.siteconfig.remove_default('valid_key_1')

        self.assertIsNone(self.siteconfig.get('valid_key_1'))

    def test_clear_defaults(self):
        """Testing SiteConfiguration.clear_defaults"""
        self.siteconfig.add_default('valid_key_1', 'valid_new_parameter_1')
        self.siteconfig.add_default('valid_key_2', 'valid_new_parameter_2')
        self.siteconfig.clear_defaults()

        self.assertIsNone(self.siteconfig.get('valid_key_1'))
        self.assertIsNone(self.siteconfig.get('valid_key_2'))

    def test_get_defaults(self):
        """Testing SiteConfiguration.get_defaults"""
        self.siteconfig.add_defaults({
            'valid_key_1': 'valid_parameter_1',
            'valid_key_2': 'valid_parameter_2',
            'valid_key_3': 'valid_parameter_3',
        })
        self.siteconfig.add_default('valid_key_1', 'valid_new_parameter_1')

        try:
            siteconfig_defaults = self.siteconfig.get_defaults()
            self.assertEqual(siteconfig_defaults['valid_key_1'],
                             'valid_new_parameter_1')
            self.assertEqual(siteconfig_defaults['valid_key_2'],
                             'valid_parameter_2')
            self.assertEqual(siteconfig_defaults['valid_key_3'],
                             'valid_parameter_3')
        finally:
            self.siteconfig.remove_default('valid_key_1')
            self.siteconfig.remove_default('valid_key_2')
            self.siteconfig.remove_default('valid_key_3')

    def test_get_defaults_excludes_global_defaults(self):
        """Testing SiteConfiguration.get_defaults excludes global defaults"""
        SiteConfiguration.add_global_default('test_global_key', 123)

        try:
            self.siteconfig.add_defaults({
                'valid_key_1': 'valid_parameter_1',
            })
            self.siteconfig.add_default('valid_key_1', 'valid_new_parameter_1')

            siteconfig_defaults = self.siteconfig.get_defaults()
            self.assertEqual(siteconfig_defaults['valid_key_1'],
                             'valid_new_parameter_1')
            self.assertNotIn('test_global_key', siteconfig_defaults)
        finally:
            SiteConfiguration.remove_global_default('test_global_key')

    def test_add_global_defaults(self):
        """Testing SiteConfiguration.add_global_defaults"""
        SiteConfiguration.add_global_defaults({
            'valid_key_1': 'global_value_1',
            'valid_key_2': 'global_value_2',
            'valid_key_3': 'global_value_3',
        })

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'global_value_1')
            self.assertEqual(self.siteconfig.get('valid_key_2'),
                             'global_value_2')
            self.assertEqual(self.siteconfig.get('valid_key_3'),
                             'global_value_3')
        finally:
            SiteConfiguration.remove_global_default('valid_key_1')
            SiteConfiguration.remove_global_default('valid_key_2')
            SiteConfiguration.remove_global_default('valid_key_3')

    def test_add_global_default(self):
        """Testing SiteConfiguration.add_global_default"""
        SiteConfiguration.add_global_default('valid_key_1', 'global_value')

        try:
            self.assertEqual(self.siteconfig.get('valid_key_1'),
                             'global_value')
        finally:
            SiteConfiguration.remove_global_default('valid_key_1')

    def test_remove_global_default(self):
        """Testing SiteConfiguration.remove_global_default"""
        SiteConfiguration.add_global_default('valid_key_1', 'global_value')
        SiteConfiguration.remove_global_default('valid_key_1')

        self.assertIsNone(self.siteconfig.get('valid_key_1'))

    def test_clear_global_defaults(self):
        """Testing SiteConfiguration.clear_global_defaults"""
        SiteConfiguration.add_global_default('valid_key_1', 'global_default_1')
        SiteConfiguration.add_global_default('valid_key_2', 'global_default_2')
        SiteConfiguration.clear_global_defaults()

        self.assertIsNone(self.siteconfig.get('valid_key_1'))
        self.assertIsNone(self.siteconfig.get('valid_key_2'))

    def test_get_global_defaults(self):
        """Testing SiteConfiguration.get_global_defaults"""
        SiteConfiguration.add_global_defaults({
            'valid_key_1': 'global_value_1',
            'valid_key_2': 'global_value_2',
            'valid_key_3': 'global_value_3',
        })
        SiteConfiguration.add_global_default('valid_key_1', 'new_global_value')

        try:
            siteconfig_defaults = SiteConfiguration.get_global_defaults()
            self.assertEqual(siteconfig_defaults['valid_key_1'],
                             'new_global_value')
            self.assertEqual(siteconfig_defaults['valid_key_2'],
                             'global_value_2')
            self.assertEqual(siteconfig_defaults['valid_key_3'],
                             'global_value_3')
        finally:
            SiteConfiguration.remove_global_default('valid_key_1')
            SiteConfiguration.remove_global_default('valid_key_2')
            SiteConfiguration.remove_global_default('valid_key_3')


class SiteConfigurationManagerTests(SiteConfigTestCase):
    """Unit tests for SiteConfigurationManager."""

    def test_check_expired_with_stale_cache(self):
        """Testing SiteConfigurationManager.check_expired with stale cache"""
        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertFalse(siteconfig1.is_expired())

        siteconfig2 = SiteConfiguration.objects.get(site=self.siteconfig.site)
        siteconfig2.set('foobar', 123)

        # Save, and prevent clearing of caches to simulate still having the
        # stale cache around for another thread.
        siteconfig2.save(clear_caches=False)

        self.assertTrue(siteconfig1.is_expired())

        SiteConfiguration.objects.check_expired()

        # See if we fetch the same one again
        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertEqual(siteconfig1.get('foobar'), 123)

    def test_check_expired_with_expired_cache(self):
        """Testing SiteConfigurationManager.check_expired with an expired
        state in cache
        """
        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertFalse(siteconfig1.is_expired())

        siteconfig2 = SiteConfiguration.objects.get(site=self.siteconfig.site)
        siteconfig2.set('foobar', 123)

        # Save, and prevent clearing of caches to simulate still having the
        # stale cache around for another thread.
        siteconfig2.save(clear_caches=False)

        cache.delete('%s:siteconfig:%s:generation' %
                     (siteconfig2.site.domain, siteconfig2.id))

        self.assertTrue(siteconfig1.is_expired())

        SiteConfiguration.objects.check_expired()

        # See if we fetch the same one again
        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertEqual(siteconfig1.get('foobar'), 123)

    def test_check_expired_emits_reloaded_signal(self):
        """Testing SiteConfigurationManager.check_expired emits
        siteconfig_reloaded when expired
        """
        signal_seen = []

        def _on_siteconfig_reloaded(siteconfig, old_siteconfig, **kwargs):
            self.assertIsNot(siteconfig, siteconfig1)
            self.assertIsNot(siteconfig, siteconfig2)
            self.assertIs(old_siteconfig, siteconfig1)
            self.assertEqual(old_siteconfig.settings, siteconfig1.settings)
            self.assertEqual(siteconfig.settings, siteconfig2.settings)
            signal_seen.append(1)

        siteconfig_reloaded.connect(_on_siteconfig_reloaded)

        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertFalse(siteconfig1.is_expired())

        siteconfig2 = SiteConfiguration.objects.get(site=self.siteconfig.site)
        siteconfig2.set('foobar', 123)

        # Save, and prevent clearing of caches to simulate still having the
        # stale cache around for another thread.
        siteconfig2.save(clear_caches=False)

        self.assertTrue(siteconfig1.is_expired())

        SiteConfiguration.objects.check_expired()

        # See if we fetch the same one again.
        siteconfig1 = SiteConfiguration.objects.get_current()
        self.assertEqual(siteconfig1.get('foobar'), 123)

        # See if the signal was emitted.
        self.assertTrue(signal_seen)



"""Base classes and mixins for new class-based generic views."""


class CheckRequestMethodViewMixin(object):
    """Generic view mixin to check the HTTP method before dispatching.

    In a normal class-based view, it's up to the
    :py:meth:`~django.views.generic.base.View.dispatch` method to ensure that
    an HTTP method is valid before dispatching to the handler. However, more
    complex views may need to perform other checks before dispatching, meaning
    that logic is being performed before the validity checks happen, which
    would be unnecessary and wasteful.

    This mixin, when put before any other mixins or parent classes, will
    perform the validity checks up-front and immediately bail with a
    :http:`405` if the HTTP method is not allowed.
    """

    def dispatch(self, request, *args, **kwargs):
        """Dispatch a HTTP request to the right handler.

        This will first check if the HTTP method is allowed. If it is not,
        then this will immediately return a :http:`405`.

        Args:
            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (dict):
                Keyword arguments to pass to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        norm_method = request.method.lower()

        if (norm_method not in self.http_method_names or
            not hasattr(self, norm_method)):
            # The HTTP method is not allowed. We can bail early.
            return self.http_method_not_allowed(request, *args, **kwargs)

        return super(CheckRequestMethodViewMixin, self).dispatch(
            request, *args, **kwargs)


class PrePostDispatchViewMixin(object):
    """Generic view mixin to call methods before/after dispatching the handler.

    This is useful for more complex generic views that take advantage of
    mixins that augment the :py:meth:`~django.views.generic.base.View.dispatch`
    method. It allows views to perform additional actions both before and
    after the HTTP handler is called.

    This mixin should be placed after any mixins that add decorators to
    :py:meth:`~django.views.generic.base.View.dispatch` and before any
    mixins/classes that inherit from :py:meth:`~django.views.generic.base.View`
    or provide an implementation of
    :py:meth:`~django.views.generic.base.View.dispatch`.

    Views should generally not have more than one version of this mixin
    applied.
    """

    def dispatch(self, *args, **kwargs):
        """Dispatch a HTTP request to the right handler.

        This will first call :py:meth:`pre_dispatch`, returning the response if
        provided by that method. It will then dispatch as normal, and then
        call :py:meth:`post_dispatch`.

        Args:
            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (dict):
                Keyword arguments to pass to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        response = self.pre_dispatch(*args, **kwargs)

        if response is not None:
            return response

        response = \
            super(PrePostDispatchViewMixin, self).dispatch(*args, **kwargs)

        return self.post_dispatch(response, *args, **kwargs)

    def pre_dispatch(self, *args, **kwargs):
        """Perform actions before dispatching to the HTTP method handler.

        This can generate state before calling the appropriate HTTP method
        handler. It can optionally return a HTTP response, which will be
        returned directly to the client.

        Subclasses don't need to call the parent method, but can.

        Args:
            *args (tuple):
                Positional arguments being passed to the handler.

            **kwargs (dict):
                Keyword arguments being passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to immediately send to the client. This
            is optional. Subclasses can return ``None`` or no value at all to
            allow the standard HTTP handler to run.
        """
        return None

    def post_dispatch(self, response, *args, **kwargs):
        """Perform actions after dispatching to the HTTP method handler.

        The handler's response is provided, allowing subclasses to alter the
        response. It must then return a response (either the existing one or a
        new one) to send to the client.

        Subclasses don't need to call the parent method, but can.

        Args:
            response (django.http.HttpResponse):
                The HTTP response generated by the handler.

            *args (tuple):
                Positional arguments that were passed to the handler.

            **kwargs (dict):
                Keyword arguments that were passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        return response



"""Compatibility imports for older extensions."""

from djblets.extensions.extension import Extension, ExtensionInfo
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.extensions.manager import ExtensionManager
from djblets.extensions.settings import ExtensionSettings, Settings


__all__ = [
    'Extension', 'ExtensionHook', 'ExtensionHookPoint', 'ExtensionInfo',
    'ExtensionManager', 'ExtensionSettings', 'Settings',
]



"""Unit tests for the ExtensionResource mixin."""

from django.test import RequestFactory

from djblets.extensions.extension import Extension
from djblets.extensions.resources import (ExtensionResource,
                                          ExtensionRootResourceMixin)
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase
from djblets.webapi.resources import RootResource, WebAPIResource


class MyTestChildResource1(WebAPIResource):
    """Test api resource item under MockResourceOne."""

    name = 'test_child_resource_item_1'
    uri_object_key = 'test_child_resource_item_1_id'


class MyTestChildResource2(WebAPIResource):
    """Test api resource item under MockResourceTwo."""

    name = 'test_child_resource_item_2'
    uri_object_key = 'test_child_resource_item_2_id'


class MyTestResource1(WebAPIResource):
    """Test api resource exposed by MockExtension."""

    name = 'test_resource_1'
    item_child_resources = [MyTestChildResource1()]


class MyTestResource2(WebAPIResource):
    """Test api resource exposed by MockExtension."""

    name = 'test_resource_2'
    item_child_resources = [MyTestChildResource2()]


class MyTestExtension(Extension):
    """Test extension for use by ExtensionTemplateTests."""

    resources = [MyTestResource1(), MyTestResource2()]


class MyTestRootResource(ExtensionRootResourceMixin, RootResource):
    """RootResource for testing ExtensionRootResourceMixin."""

    def get_extension_resource(self):
        """Return the mocked extension resource.

        Returns:
            djblets.extensions.resources.ExtensionResource:
            The mock extension resource.
        """
        return self.list_child_resources[0]


class ExtensionRootResourceMixinTests(ExtensionTestCaseMixin, TestCase):
    """Unit tests for the ExtensionRootResourceMixin."""

    extension_class = MyTestExtension

    def setUp(self):
        super(ExtensionRootResourceMixinTests, self).setUp()

        self.ext_resource = ExtensionResource(self.extension_mgr)
        self.root_resource = MyTestRootResource([self.ext_resource])
        self.request = RequestFactory().get('/')

    def test_generate_extension_uris_for_template(self):
        """Testing ExtensionRootResourceMixin generates URI templates when
        extensions are initialized
        """
        self.root_resource._generate_extension_uris_for_template(
            self.extension)
        actual_result = self.root_resource.get_uri_templates(
            self.request)
        self.assertEqual(actual_result, {
            'extensions': 'http://testserver/extensions/',
            'extension': 'http://testserver/extensions/{extension_name}/',
            'test_resource_1s': (
                'http://testserver/extensions/djblets.webapi.tests.'
                'test_extension_resource_mixin.'
                'MyTestExtension/test-resource-1s/'
            ),
            'test_resource_2s': (
                'http://testserver/extensions/djblets.webapi.tests.'
                'test_extension_resource_mixin.'
                'MyTestExtension/test-resource-2s/'
            ),
            'root': 'http://testserver/',
        })

    def test_remove_extension_uris_from_template(self):
        """Testing ExtensionRootResourceMixin removes URI templates when
        extensions are disabled
        """
        self.root_resource._generate_extension_uris_for_template(
            self.extension)
        self.root_resource._remove_extension_uris_from_template(
            self.extension)
        actual_result = self.root_resource.get_uri_templates(
            self.request)
        self.assertEqual(actual_result, {
            'extensions': u'http://testserver/extensions/',
            'root': 'http://testserver/',
            'extension': 'http://testserver/extensions/{extension_name}/',
        })



"""Unit tests for djblets.privacy.consent.views."""

from urllib.parse import quote

from django.contrib.auth.models import AnonymousUser, User
from django.core.cache import cache
from django.http import HttpResponse, HttpResponseRedirect
from django.test.client import RequestFactory
from django.test.utils import override_settings
from django.views.generic.base import View

from djblets.privacy.consent import (BaseConsentRequirement,
                                     ConsentData,
                                     get_consent_requirements_registry,
                                     get_consent_tracker)
from djblets.privacy.consent.common import PolicyConsentRequirement
from djblets.privacy.consent.tracker import clear_consent_tracker
from djblets.privacy.consent.views import (CheckPendingConsentMixin,
                                           check_pending_consent)
from djblets.testing.testcases import TestCase
from djblets.views.generic.base import PrePostDispatchViewMixin


class NefariousConsentRequirement(BaseConsentRequirement):
    """A consent requirement for testing."""

    requirement_id = 'nefarious'
    name = 'Nefarious Requirement'
    intent_description = 'We want your data'
    data_use_description = 'We want to use your data for nefarious purposes.'


class BenevolentConsentRequirement(BaseConsentRequirement):
    """A consent requirement for testing."""

    requirement_id = 'Benevolent'
    name = 'Some Requirement'
    intent_description = 'We want your data'
    data_use_description = 'We want to use your data for benevolent purposes.'


class MixinView(CheckPendingConsentMixin, PrePostDispatchViewMixin, View):
    """A view for testing."""

    def get(self, request):
        return HttpResponse(b'ok', status=200)


@check_pending_consent
def decorated_view(request):
    """A view for testing."""
    return HttpResponse(b'ok', status=200)


@override_settings(DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL='/consent')
class CheckPendingConsentTests(TestCase):
    """Unit tests for djblets.privacy.consent.views."""

    @classmethod
    def setUpClass(cls):
        super(CheckPendingConsentTests, cls).setUpClass()

        cls.request_factory = RequestFactory()
        cls.registry = get_consent_requirements_registry()

    def setUp(self):
        super(CheckPendingConsentTests, self).setUp()

        self.registry.register(NefariousConsentRequirement())
        self.registry.register(BenevolentConsentRequirement())

    def tearDown(self):
        super(CheckPendingConsentTests, self).tearDown()

        self.registry.reset()
        clear_consent_tracker()
        cache.clear()

    def test_decorator_anonymous(self):
        """Testing @check_pending_consent when the user is anonymous"""
        request = self.request_factory.get('/')
        request.user = AnonymousUser()

        rsp = decorated_view(request)

        self.assertIsInstance(rsp, HttpResponse)
        self.assertEqual(rsp.status_code, 200)
        self.assertEqual(rsp.content, b'ok')

    def test_decorator_all_met(self):
        """Testing @check_pending_consent when a user has no pending consent
        decisions
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        tracker = get_consent_tracker()
        tracker.record_consent_data_list(
            request.user,
            [
                ConsentData(BenevolentConsentRequirement.requirement_id,
                            granted=True),
                ConsentData(NefariousConsentRequirement.requirement_id,
                            granted=False),
            ])

        rsp = decorated_view(request)

        self.assertNotIsInstance(rsp, HttpResponseRedirect)
        self.assertIsInstance(rsp, HttpResponse)
        self.assertEqual(rsp.content, b'ok')

    def test_decorator_some_met(self):
        """Testing @check_pending_consent when a user has some pending consent
        decisions
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        get_consent_tracker().record_consent_data(
            request.user,
            ConsentData(BenevolentConsentRequirement.requirement_id,
                        granted=True))

        rsp = decorated_view(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')

    def test_decorator_none_met(self):
        """Testing @check_pending_consent when a user has all consent
        decisions pending
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        rsp = decorated_view(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')

    def test_decorator_function_url(self):
        """Testing @check_pending_consent when
        DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL is a function
        """
        def redirect_url(request):
            return '/consent?next=%s' % quote(request.get_full_path())

        request = self.request_factory.get('/foo/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        with self.settings(
            DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL=redirect_url):
            rsp = decorated_view(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent?next=/foo/')

    def test_decorator_reject_policy(self):
        """Testing @check_pending_consent when the user has rejected the policy
        requirement
        """
        policy_requirement = PolicyConsentRequirement(
            'https://example.com/',
            'https://example.com/',
            reject_instructions='Obey.')

        self.registry.register(policy_requirement)

        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        get_consent_tracker().record_consent_data_list(
            request.user,
            [
                ConsentData(BenevolentConsentRequirement.requirement_id,
                            granted=True),
                ConsentData(NefariousConsentRequirement.requirement_id,
                            granted=False),
                policy_requirement.build_consent_data(granted=False),
            ])

        rsp = decorated_view(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')

    def test_mixin_anonymous(self):
        """Testing CheckPendingConsentMixin when the user is anonymous"""
        request = self.request_factory.get('/')
        request.user = AnonymousUser()

        rsp = MixinView.as_view()(request)

        self.assertIsInstance(rsp, HttpResponse)
        self.assertEqual(rsp.status_code, 200)
        self.assertEqual(rsp.content, b'ok')

    def test_mixin_all_met(self):
        """Testing CheckPendingConsentMixin when a user has no pending consent
        decisions
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        get_consent_tracker().record_consent_data_list(
            request.user,
            [
                ConsentData(BenevolentConsentRequirement.requirement_id,
                            granted=True),
                ConsentData(NefariousConsentRequirement.requirement_id,
                            granted=False),
            ])

        rsp = MixinView.as_view()(request)

        self.assertNotIsInstance(rsp, HttpResponseRedirect)
        self.assertIsInstance(rsp, HttpResponse)
        self.assertEqual(rsp.content, b'ok')

    def test_mixin_some_met(self):
        """Testing CheckPendingConsentMixin when a user has some pending
        consent decisions
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(
            username='user', email='user@example.com')

        get_consent_tracker().record_consent_data(
            request.user,
            ConsentData(BenevolentConsentRequirement.requirement_id,
                        granted=True))

        rsp = MixinView.as_view()(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')

    def test_mixin_none_met(self):
        """Testing CheckPendingConsentMixin when a user has all consent
        decisions pending
        """
        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        rsp = MixinView.as_view()(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')

    def test_mixin_function_url(self):
        """Testing CheckPendingConsentMixin when
        DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL is a function
        """
        def redirect_url(request):
            return '/consent?next=%s' % quote(request.get_full_path())

        request = self.request_factory.get('/foo/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        with self.settings(
            DJBLETS_PRIVACY_PENDING_CONSENT_REDIRECT_URL=redirect_url):
            rsp = MixinView.as_view()(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent?next=/foo/')

    def test_mixin_reject_policy(self):
        """Testing CheckPendingConsentMixin when the user has rejected the
        policy requirement
        """
        policy_requirement = PolicyConsentRequirement(
            'https://example.com/',
            'https://example.com/',
            reject_instructions='Obey.')

        self.registry.register(policy_requirement)

        request = self.request_factory.get('/')
        request.user = User.objects.create_user(username='user',
                                                email='user@example.com')

        get_consent_tracker().record_consent_data_list(
            request.user,
            [
                ConsentData(BenevolentConsentRequirement.requirement_id,
                            granted=True),
                ConsentData(NefariousConsentRequirement.requirement_id,
                            granted=False),
                policy_requirement.build_consent_data(granted=False),
            ])

        rsp = MixinView.as_view()(request)

        self.assertIsInstance(rsp, HttpResponseRedirect)
        self.assertEqual(rsp.url, '/consent')



"""Base support and standard value field wrappers for conditions."""

import re

from django import forms
from django.utils.translation import gettext_lazy as _

from djblets.conditions.errors import InvalidConditionValueError


class BaseConditionValueField(object):
    """Base class for a field for editing and representing condition values.

    This is used to provide a field in the UI that can be used for editing a
    condition value. It's responsible for rendering the field, preparing
    data for the field, retrieving the data from the HTML form data, and
    handling JSON-safe serialization/deserialization of values.

    Subclasses can specify custom logic for all these operations, and can
    specify the JavaScript counterparts for the class used to edit the values.
    """

    #: The JavaScript model class for representing field state.
    #:
    #: This is instantiated on the web UI and is used to store any model
    #: data provided by :py:meth:`get_js_model_data`.
    #:
    #: The default is a simple model that just stores the model data as
    #: attributes.
    js_model_class = 'Djblets.Forms.ConditionValueField'

    #: The JavaScript view class for editing fields.
    #:
    #: This is instantiated on the web UI and is used to provide an editor
    #: for the condition's value.
    #:
    #: It's passed any options that are returned from
    #: :py:meth:`get_js_model_data`.
    js_view_class = None

    def serialize_value(self, value):
        """Serialize a Python object into a JSON-compatible serialized form.

        This is responsible for taking a Python value/object of some sort
        (string, list, or anything more complex) and returning a
        JSON-compatible form for serialization.

        By default, this returns the value as-is.

        Args:
            value (object):
                The value to serialize.

        Returns:
            object:
            The JSON-compatible serialized value.
        """
        return value

    def deserialize_value(self, serialized_value):
        """Deserialize a value back into a Python object.

        This is responsible for taking a value serialized by
        :py:meth:`serialize_value` and returning a suitable Python
        object/value.

        By default, this returns the value as-is.

        Args:
            serialized_value (object):
                The serialized value to deserialize.

        Returns:
            object:
            The deserialized value.

        Raises:
            djblets.conditions.errors.InvalidConditionValueError:
                Error deserializing or validating the data.
        """
        return serialized_value

    def get_from_form_data(self, data, files, name):
        """Return a value from a form data dictionary.

        This attempts to return the value for a condition from Django form
        data. It's passed a dictionary of data, uploaded files, and the name
        of the appropriate value field.

        Subclasses can override this to normalize the value before returning.

        Args:
            data (django.http.request.QueryDict):
                The dictionary containing form data.

            files (django.http.request.QueryDict):
                The dictionary containing uploaded files.

            name (unicode):
                The field name for the value to load.

        Returns:
            object:
            The value from the form data.
        """
        return data.get(name, None)

    def prepare_value_for_widget(self, value):
        """Return a value suitable for use in the widget.

        The value will be passed to the widget's JavaScript UI. It can be
        used in special cases where a Python object needs to be converted
        to another form in order to work properly client-side.

        By default, the value is returned as-is.

        Args:
            value (object):
                The value to prepare for the widget.

        Returns:
            object:
            The value prepared for the widget.
        """
        return value

    def get_js_model_data(self):
        """Return data for the JavaScript model for this field.

        The returned data will be set as attributes on the Backbone model
        pointed to by :py:attr:`js_model_class`.

        By default, this includes the rendered HTML as ``fieldHTML``, which
        should generally be provided (but is not required, depending on the
        field).

        Returns:
            dict:
            The model data. This must be serializable as JSON.
        """
        return {
            'fieldHTML': self.render_html(),
        }

    def get_js_view_data(self):
        """Return data for the JavaScript view for this field.

        The returned data will be set as options on the Backbone view pointed
        to by :py:attr:`js_view_class`.

        This is empty by default.

        Returns:
            dict:
            The view data. This must be serializable as JSON.
        """
        return {}

    def render_html(self):
        """Return rendered HTML for the field.

        The rendered HTML will be inserted dynamically by the JavaScript UI.

        This must be implemented by subclasses.

        Returns:
            unicode:
            The rendered HTML for the field. This does not need to be marked as
            safe (but can be), as it will be passed in as an escaped JavaScript
            string.
        """
        raise NotImplementedError


class ConditionValueFormField(BaseConditionValueField):
    """Condition value wrapper for HTML form fields.

    This allows the usage of standard HTML form fields (through Django's
    :py:mod:`django.forms` module) for rendering and accepting condition
    values.

    Callers simply need to instantiate the class along with a form field.

    The rendered field must support setting and getting a ``value``
    attribute on the DOM element, like a standard HTML form field.

    Example:
        value_field = ConditionValueFormField(
            forms.ModelMultipleChoiceField(queryset=MyModel.objects.all()))
    """

    js_model_class = 'Djblets.Forms.ConditionValueField'
    js_view_class = 'Djblets.Forms.ConditionValueFormFieldView'

    def __init__(self, field):
        """Initialize the value field.

        Args:
            field (django.forms.fields.Field):
                The Django form field instance for the value. This may also
                be a callable that returns a field.
        """
        super(ConditionValueFormField, self).__init__()

        self.field = field

    @property
    def field(self):
        """The form field to use for the value.

        This will always return a :py:class:`~django.forms.fields.Field`,
        but can be given a callable that returns a field when set.
        """
        if callable(self._field):
            self._field = self._field()

        return self._field

    # Note that the docstring will be inherited from the field property.
    @field.setter
    def field(self, field):
        self._field = field

    def serialize_value(self, value):
        """Serialize a Python object into a JSON-compatible serialized form.

        This is responsible for taking a Python value/object of some sort
        (string, list, or anything more complex) and returning a
        JSON-compatible form for serialization. It will use the form field
        to do this (through :py:meth:`Field.prepare_value()
        <django.forms.fields.Field.prepare_value>`).

        Args:
            value (object):
                The value to serialize.

        Returns:
            object:
            The JSON-compatible serialized value.
        """
        return self.field.prepare_value(value)

    def deserialize_value(self, value_data):
        """Deserialize a value back into a Python object.

        This is responsible for taking a value serialized by
        :py:meth:`serialize_value` and returning a suitable Python
        object/value. It will use the form field to do this (through
        :py:meth:`Field.clean() <django.forms.fields.Field.clean>`).

        By default, this returns the value as-is.

        Args:
            serialized_value (object):
                The serialized value to deserialize.

        Returns:
            object:
            The deserialized value.

        Raises:
            djblets.conditions.errors.InvalidConditionValueError:
                Error deserializing or validating the data.
        """
        try:
            return self.field.clean(value_data)
        except forms.ValidationError as e:
            raise InvalidConditionValueError('; '.join(e.messages),
                                             code=e.code)

    def get_from_form_data(self, data, files, name):
        """Return a value from a form data dictionary.

        This attempts to return the value for a condition from Django form
        data. It's passed a dictionary of data, uploaded files, and the name
        of the appropriate value field. It will use the form field's widget to
        do this (through :py:meth:`Widget.value_from_datadict
        <django.forms.widgets.Widget.value_from_datadict>`).

        Args:
            data (django.http.request.QueryDict):
                The dictionary containing form data.

            files (django.http.request.QueryDict):
                The dictionary containing uploaded files.

            name (unicode):
                The field name for the value to load.

        Returns:
            object:
            The value from the form data.
        """
        return self.field.widget.value_from_datadict(data, files, name)

    def render_html(self):
        """Return rendered HTML for the field.

        The rendered HTML will be generated by the widget for the field,
        and will be dynamically inserted by the JavaScript UI.

        Returns:
            unicode:
            The rendered HTML for the field.
        """
        # The name is a placeholder, and will be updated by the JavaScript.
        # However, we must have it for render.
        return self.field.widget.render(name='XXX', value=None)


class ConditionValueBooleanField(ConditionValueFormField):
    """Condition value wrapper for boolean form fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.BooleanField`. It accepts the same
    keyword arguments in the constructor that the field itself accepts.

    It also specially serializes the value to a string for use in the
    JavaScript widget.

    Example:
        value_field = ConditionValueBooleanField(initial=True)
    """

    def __init__(self, **field_kwargs):
        """Initialize the value field.

        Args:
            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.BooleanField` constructor.
        """
        widget = field_kwargs.pop('widget', None)

        if not widget:
            widget = forms.widgets.Select(choices=(
                (True, _('True')),
                (False, _('False')),
            ))

        super(ConditionValueBooleanField, self).__init__(
            field=forms.BooleanField(
                required=False,
                widget=widget,
                **field_kwargs))

    def prepare_value_for_widget(self, value):
        """Return a value suitable for use in the widget.

        This will convert a boolean value to a string, so that it can be
        properly matched against the string choices for the select box.

        Args:
            value (bool):
                The value to prepare for the widget.

        Returns:
            unicode:
            The string value for the widget.
        """
        if value:
            return 'True'
        else:
            return 'False'


class ConditionValueCharField(ConditionValueFormField):
    """Condition value wrapper for single-line text form fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.CharField`. It accepts the same keyword
    arguments in the constructor that the field itself accepts.

    Example:
        value_field = ConditionValueCharField(max_length=100)
    """

    def __init__(self, **field_kwargs):
        """Initialize the value field.

        Args:
            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.CharField` constructor.
        """
        super(ConditionValueCharField, self).__init__(
            field=forms.CharField(**field_kwargs))


class ConditionValueIntegerField(ConditionValueFormField):
    """Condition value wrapper for integer form fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.IntegerField`. It accepts the same
    keyword arguments in the constructor that the field itself accepts.

    Example:
        value_field = ConditionValueIntegerField(min_value=0, max_value=100)
    """

    def __init__(self, **field_kwargs):
        """Initialize the value field.

        Args:
            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.IntegerField` constructor.
        """
        super(ConditionValueIntegerField, self).__init__(
            field=forms.IntegerField(**field_kwargs))


class ConditionValueMultipleChoiceField(ConditionValueFormField):
    """Condition value wrapper for multiple choice fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.MultipleChoiceField`. It accepts the same
    keyword arguments in the constructor that the field itself accepts.

    Example:
        value_field = ConditionValueMultipleChoiceField(
            choices=[
                ('value1', 'Value 1'),
                ('value2', 'Value 2'),
            ])

    Version Added:
        3.0
    """

    def __init__(self, **field_kwargs):
        """Initialize the value field.

        Args:
            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.MultipleChoiceField`
                constructor.
        """
        super().__init__(field=forms.MultipleChoiceField(**field_kwargs))


class ConditionValueModelField(ConditionValueFormField):
    """Condition value wrapper for single model form fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.ModelChoiceField`. It accepts the same
    keyword arguments in the constructor that the field itself accepts.

    Unlike the standard field, the provided queryset can be a callable that
    returns a queryset.

    Example:
        value_field = ConditionValueModelField(queryset=MyObject.objects.all())
    """

    def __init__(self, queryset, **field_kwargs):
        """Initialize the value field.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset used for the field. This may also be a callable
                that returns a queryset.

            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.ModelChoiceField` constructor.
        """
        def _build_field():
            if callable(queryset):
                qs = queryset()
            else:
                qs = queryset

            empty_label = field_kwargs.pop('empty_label', None)

            return forms.ModelChoiceField(queryset=qs,
                                          empty_label=empty_label,
                                          **field_kwargs)

        super(ConditionValueModelField, self).__init__(field=_build_field)


class ConditionValueMultipleModelField(ConditionValueFormField):
    """Condition value wrapper for multiple model form fields.

    This is a convenience for condition values that want to use a
    :py:class:`~django.forms.fields.ModelMutipleChoiceField`. It accepts the
    same keyword arguments in the constructor that the field itself accepts.

    Unlike the standard field, the provided queryset can be a callable that
    returns a queryset.

    Example:
        value_field = ConditionValueMultipleModelField(
            queryset=MyObject.objects.all())
    """

    def __init__(self, queryset, **field_kwargs):
        """Initialize the value field.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset used for the field. This may also be a callable
                that returns a queryset.

            **field_kwargs (dict):
                Keyword arguments to pass to the
                :py:class:`~django.forms.fields.ModelChoiceField` constructor.
        """
        def _build_field():
            if callable(queryset):
                qs = queryset()
            else:
                qs = queryset

            return forms.ModelMultipleChoiceField(queryset=qs, **field_kwargs)

        super(ConditionValueMultipleModelField, self).__init__(
            field=_build_field)


class ConditionValueRegexField(ConditionValueCharField):
    """Condition value for fields that accept regexes.

    This value accepts and validates regex patterns entered into the field.

    Example:
        value_field = ConditionValueRegexField()
    """

    def serialize_value(self, value):
        """Serialize a compiled regex into a string.

        Args:
            value (object):
                The value to serialize.

        Returns:
            object:
            The JSON-compatible serialized value.
        """
        return value.pattern

    def deserialize_value(self, value_data):
        """Deserialize a regex pattern string into a compiled regex.

        Args:
            value_data (unicode):
                The serialized regex pattern to compile.

        Returns:
            object:
            The deserialized value.

        Raises:
            djblets.conditions.errors.InvalidConditionValueError:
                The regex could not be compiled.
        """
        try:
            return re.compile(value_data, re.UNICODE)
        except re.error as e:
            raise InvalidConditionValueError(
                'Your regex pattern had an error: %s' % e)



"""A registry for configuration forms and pages."""

from django.utils.translation import gettext_lazy as _

from djblets.registries.errors import RegistrationError
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED, DEFAULT_ERRORS,
                                         OrderedRegistry, UNREGISTER)


# Errors for ConfigPageFormRegistry
CONFIG_PAGE_FORM_REGISTRY_DEFAULT_ERRORS = DEFAULT_ERRORS.copy()
CONFIG_PAGE_FORM_REGISTRY_DEFAULT_ERRORS.update({
    ALREADY_REGISTERED: _(
        'Could not register form %(item)r: This form is already '
        'registered.'
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register form %(item)r: Another form (%(duplicate)r) '
        'is already registered with %(attr_name)s = %(attr_value)s.'
    ),
})


FORM_ALREADY_REGISTERED = 'form_already_registered'

# Errors for ConfigPageRegistry
CONFIG_PAGE_REGISTRY_DEFAULT_ERRORS = DEFAULT_ERRORS.copy()
CONFIG_PAGE_REGISTRY_DEFAULT_ERRORS.update({
    ALREADY_REGISTERED: _(
        'Could not register page %(item)s: This page is already registered.'
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register page %(item)s: Another page (%(duplicate)s) '
        'is already registered with %(attr_name)s = %(attr_value)s.'
    ),
    FORM_ALREADY_REGISTERED: _(
        'Could not register page %(page)s: One of its forms (%(form)s) has '
        'already been registered with the %(duplicate)s page.',
    ),
    UNREGISTER: _(
        'Failed to unregister unknown account page %(item)s: This page is is '
        'not registered.'
    ),
})


class ConfigPageFormRegistry(OrderedRegistry):
    """A registry for managing configuration page forms."""

    lookup_attrs = ('form_id',)
    default_errors = CONFIG_PAGE_FORM_REGISTRY_DEFAULT_ERRORS


class ConfigPageRegistry(OrderedRegistry):
    """A registry for managing configuration pages.

    This allows subclasses to dynamically change which pages are available, as
    well as which forms are available on each page.

    The pages managed by this registry must subclass
    :py:class:`~djblets.configforms.mixins.DynamicConfigPageMixin`.
    """

    lookup_attrs = ('page_id',)
    default_errors = CONFIG_PAGE_REGISTRY_DEFAULT_ERRORS

    def __init__(self):
        super(ConfigPageRegistry, self).__init__()

        self._forms = ConfigPageFormRegistry()

        # Map form IDs to page IDs so that we can find pages that have already
        # registered forms when errors occur.
        self._form_to_page = {}

    def register(self, page_class):
        """Register a configuration page class.

        A page ID is considered unique and can only be registered once.

        This will also register all form classes on the page. If registration
        for any form fails, registration for the entire class will fail. In
        this case, the page will become unregistered, as well as any forms on
        that page that were successfully registered.

        Args:
            page_class (type):
                The page class to register, as a subclass of
                :py:class:`djblets.configforms.pages.ConfigPage`.

        Raises:
            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the page has already been registered.

            djblets.registries.errors.RegistrationError:
                Raised if the page shares an attribute with an already
                registered page or if any of its forms share an attribute
                with an already registered form.
        """
        super(ConfigPageRegistry, self).register(page_class)

        # Set the form_classes to an empty list by default if it doesn't
        # explicitly provide its own, so that entries don't go into
        # DynamicConfigurationPageMixin's global list.
        if page_class.form_classes is None:
            page_class.form_classes = []

        # Set _default_form_classes when an account page class first registers.
        if page_class._default_form_classes is None:
            page_class._default_form_classes = list(page_class.form_classes)

        # If form_classes is empty, reload the list from _default_form_classes.
        if not page_class.form_classes:
            page_class.form_classes = list(page_class._default_form_classes)

        # We keep track of the pages that are successfully registered so that
        # if registration of any page fails, we can rollback those during
        # page unregistration.
        registered_forms = []

        for form_class in page_class.form_classes:
            try:
                self._forms.register(form_class)
                self._form_to_page[form_class.form_id] = page_class.page_id
            except RegistrationError:
                duplicate = self.get('page_id',
                                     self._form_to_page[form_class.form_id])

                self.unregister(page_class, registered_forms)
                raise RegistrationError(self.format_error(
                    FORM_ALREADY_REGISTERED,
                    page=page_class,
                    form=form_class,
                    duplicate=duplicate))

            registered_forms.append(form_class)

    def unregister(self, page_class, registered_forms=None):
        """Unregister a configuration page class.

        Args:
            page_class (type):
                The page class to unregister, as a subclass of
                :py:class:`djblets.configforms.pages.ConfigPage`.

            registered_forms (list, optional):
                The forms that have been successfully registered for the class.
                This attribute is only used when registration of forms was only
                partially successful.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised when the specified class or form is not registered.
        """
        super(ConfigPageRegistry, self).unregister(page_class)

        if registered_forms is None:
            # We copy the data so that we are not modifying the list while
            # iterating through it.
            registered_forms = list(page_class.form_classes)

        for form_class in registered_forms:
            self.remove_form_from_page(page_class, form_class)

    def add_form_to_page(self, page_class, form_class):
        """Add a form to the page.

        Callers should prefer to use
        :py:meth:`~djblets.configforms.mixins.DynamicConfigPageMixin.add_form`
        over this method.

        Args:
            page_class (type):
                The page to add the form to, as a subclass of
                :py:class:`djblets.configforms.pages.ConfigPage`.

            form_class (type):
                The form to add to the page, as a subclass of
                :py:class:`djblets.configforms.page.ConfigPage`.

        Raises:
            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the form has already been registered.

            djblets.registries.errors.RegistrationError:
                Raised if the form shares an attribute with an already
                registered form.
        """
        self.populate()
        self._forms.register(form_class)
        page_class.form_classes.append(form_class)
        self._form_to_page[form_class.form_id] = page_class.page_id

    def remove_form_from_page(self, page_class, form_class):
        """Remove a form from the page.

        Callers should prefer to use
        :py:meth:`~djblets.configforms.mixins.DynamicConfigPageMixin.remove_form`
        over this method.

        Args:
            page_class (type):
                The page to remove the form from, as a subclass of
                :py:class:`djblets.configforms.pages.ConfigPage`.

            form_class (type):
                The form to remove from the page, as a subclass of
                :py:class:`djblets.configforms.forms.ConfigPageForm`.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the form was not previously registered.
        """
        self.populate()
        self._forms.unregister(form_class)
        page_class.form_classes.remove(form_class)
        del self._form_to_page[form_class.form_id]



"""Unit tests for djblets.db.fields.base64_field."""

from django.db import connection, models

from djblets.db.fields.base64_field import (Base64DecodedValue, Base64Field,
                                            Base64TypeError)
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class Base64TestModel(models.Model):
    field = Base64Field(null=True)


class Base64FieldTests(TestModelsLoaderMixin, TestCase):
    """Unit tests for djblets.db.fields.base64_field.Base64Field."""

    tests_app = 'djblets.db.tests'

    def test_create_with_decoded_bytes_value(self):
        """Testing Base64Field with setting decoded bytes value in
        Model.objects.create()
        """
        obj = Base64TestModel.objects.create(field=b'This is a test')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_create_with_decoded_unicode_value(self):
        """Testing Base64Field with setting decoded unicode value in
        Model.objects.create()
        """
        obj = Base64TestModel.objects.create(field='This is a tést')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_create_with_base64_decoded_value(self):
        """Testing Base64Field with setting Base64DecodedValue in
        Model.objects.create()
        """
        obj = Base64TestModel.objects.create(
            field=Base64DecodedValue(b'This is a test'))

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_create_with_memoryview(self):
        """Testing Base64Field with setting memoryview in
        Model.objects.create()
        """
        obj = Base64TestModel.objects.create(
            field=memoryview(b'This is a t\xc3\xa9st'))

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_create_with_none(self):
        """Testing Base64Field with setting None in
        Model.objects.create()
        """
        obj = Base64TestModel.objects.create(field=None)
        self.assertIsNone(obj.field)

        encoded = obj.get_field_base64()
        self.assertIsNone(encoded)

    def test_create_with_invalid_value_type(self):
        """Testing Base64Field with setting invalid value type in
        Model.objects.create()
        """
        with self.assertRaises(Base64TypeError):
            Base64TestModel.objects.create(field=True)

    def test_unsaved_obj_with_decoded_bytes_value(self):
        """Testing Base64Field with setting decoded bytes value on unsaved
        instance
        """
        obj = Base64TestModel()
        obj.field = b'This is a test'

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_unsaved_obj_with_decoded_unicode_value(self):
        """Testing Base64Field with setting decoded unicode value on unsaved
        instance
        """
        obj = Base64TestModel()
        obj.field = 'This is a tést'

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_unsaved_obj_with_base64_decoded_value(self):
        """Testing Base64Field with setting Base64DecodedValue on unsaved
        instance
        """
        obj = Base64TestModel()
        obj.field = Base64DecodedValue(b'This is a test')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_unsaved_obj_with_memoryview_value(self):
        """Testing Base64Field with setting memoryview on unsaved instance"""
        obj = Base64TestModel()
        obj.field = memoryview(b'This is a t\xc3\xa9st')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_unsaved_obj_with_invalid_value_type(self):
        """Testing Base64Field with setting invalid value type on unsaved
        instance
        """
        obj = Base64TestModel()

        with self.assertRaises(Base64TypeError):
            obj.field = True

    def test_unsaved_obj_with_none(self):
        """Testing Base64Field with setting None on unsaved instance"""
        obj = Base64TestModel()
        obj.field = None

        self.assertIsNone(obj.field)

        encoded = obj.get_field_base64()
        self.assertIsNone(encoded)

    def test_saved_obj_with_decoded_bytes_value(self):
        """Testing Base64Field with setting decoded bytes value on saved
        instance
        """
        obj = Base64TestModel.objects.create()
        obj.field = b'VGhpcyBpcyBhIHRlc3Q=\n'

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_saved_obj_with_decoded_unicode_value(self):
        """Testing Base64Field with setting decoded unicode value on saved
        instance
        """
        obj = Base64TestModel.objects.create()
        obj.field = 'VGhpcyBpcyBhIHTDqXN0\n'

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_saved_obj_with_base64_decoded_value(self):
        """Testing Base64Field with setting Base64DecodedValue on saved
        instance
        """
        obj = Base64TestModel.objects.create()
        obj.field = Base64DecodedValue(b'This is a test')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_saved_obj_with_memoryview_value(self):
        """Testing Base64Field with setting memoryview on saved instance"""
        obj = Base64TestModel.objects.create()
        obj.field = memoryview(b'VGhpcyBpcyBhIHTDqXN0\n')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_saved_obj_with_invalid_value_type(self):
        """Testing Base64Field with setting invalid value type on saved
        instance
        """
        obj = Base64TestModel.objects.create()

        with self.assertRaises(Base64TypeError):
            obj.field = True

    def test_saved_obj_with_none(self):
        """Testing Base64Field with setting None on saved instance"""
        obj = Base64TestModel.objects.create()
        obj.field = None

        self.assertIsNone(obj.field)

        encoded = obj.get_field_base64()
        self.assertIsNone(encoded)

    def test_values_persist_after_save(self):
        """Testing Base64Field value persists correctly after save"""
        obj = Base64TestModel.objects.create(field=b'This is a test')
        obj = Base64TestModel.objects.get(pk=obj.pk)

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_values_persist_after_save_with_bad_x(self):
        """Testing Base64Field value persists correctly after save with
        bad \\x from Djblets 2.0
        """
        # Djblets 2.0 could save an empty Base64-encoded string as "\x", due to
        # saving as bytes. This was an issue only on Postgres, and didn't
        # manifest on MySQL or SQLite. Still, to work around it, we now check
        # for this bad value.
        obj = Base64TestModel.objects.create()

        # Simulate the error case.
        with connection.cursor() as cursor:
            # We want to plug in the table name, so we won't be using
            # execute's own params=[...] here.
            cursor.execute('UPDATE %s SET field="\\x" WHERE id=%s'
                           % (Base64TestModel._meta.db_table, obj.pk))

        obj = Base64TestModel.objects.get(pk=obj.pk)

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'')

    def test_save_form_data_with_bytes(self):
        """Testing Base64Field.save_form_data with bytes value"""
        obj = Base64TestModel(field=b'This is a test')
        obj._meta.get_field('field').save_form_data(obj, b'This is a test')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a test')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_save_form_data_with_unicode(self):
        """Testing Base64Field.save_form_data with unicode value"""
        obj = Base64TestModel(field=b'This is a test')
        obj._meta.get_field('field').save_form_data(obj, 'This is a tést')

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_save_form_data_with_memoryview(self):
        """Testing Base64Field.save_form_data with memoryview value"""
        obj = Base64TestModel(field=b'This is a test')
        obj._meta.get_field('field').save_form_data(
            obj, memoryview(b'This is a t\xc3\xa9st'))

        self.assertIs(type(obj.field), Base64DecodedValue)
        self.assertEqual(obj.field, b'This is a t\xc3\xa9st')

        encoded = obj.get_field_base64()
        self.assertIs(type(encoded), bytes)
        self.assertEqual(encoded, b'VGhpcyBpcyBhIHTDqXN0\n')

    def test_get_prep_value_with_bytes(self):
        """Testing Base64Field.get_prep_value with bytes value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').get_prep_value(
            b'VGhpcyBpcyBhIHTDqXN0\n')

        self.assertIs(type(value), str)
        self.assertEqual(value, 'VGhpcyBpcyBhIHTDqXN0\n')

    def test_get_prep_value_with_unicode(self):
        """Testing Base64Field.get_prep_value with unicode value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').get_prep_value(
            'VGhpcyBpcyBhIHTDqXN0\n')

        self.assertIs(type(value), str)
        self.assertEqual(value, 'VGhpcyBpcyBhIHTDqXN0\n')

    def test_get_prep_value_with_base64_decoded_value(self):
        """Testing Base64Field.get_prep_value with Base64DecodedValue value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').get_prep_value(
            Base64DecodedValue(b'This is a t\xc3\xa9st'))

        self.assertIs(type(value), str)
        self.assertEqual(value, 'VGhpcyBpcyBhIHTDqXN0\n')

    def test_get_prep_value_with_memoryview_value(self):
        """Testing Base64Field.get_prep_value with memoryview value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').get_prep_value(
            memoryview(b'VGhpcyBpcyBhIHTDqXN0\n'))

        self.assertIs(type(value), str)
        self.assertEqual(value, 'VGhpcyBpcyBhIHTDqXN0\n')

    def test_get_prep_value_with_none(self):
        """Testing Base64Field.get_prep_value with None value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').get_prep_value(None)

        self.assertIsNone(value)

    def test_to_python_with_bytes(self):
        """Testing Base64Field.to_python with bytes value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').to_python(
            b'VGhpcyBpcyBhIHRlc3Q=\n')

        self.assertIs(type(value), Base64DecodedValue)
        self.assertEqual(value, b'This is a test')

    def test_to_python_with_unicode(self):
        """Testing Base64Field.to_python with unicode value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').to_python(
            'VGhpcyBpcyBhIHRlc3Q=\n')

        self.assertIs(type(value), Base64DecodedValue)
        self.assertEqual(value, b'This is a test')

    def test_to_python_with_base64_decoded_value(self):
        """Testing Base64Field.to_python with Base64DecodedValue"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').to_python(
            Base64DecodedValue(b'This is a test'))

        self.assertIs(type(value), Base64DecodedValue)
        self.assertEqual(value, b'This is a test')

    def test_to_python_with_memoryview_value(self):
        """Testing Base64Field.to_python with memoryview value"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').to_python(
            memoryview(b'VGhpcyBpcyBhIHTDqXN0\n'))

        self.assertIs(type(value), Base64DecodedValue)
        self.assertEqual(value, b'This is a t\xc3\xa9st')

    def test_to_python_with_none(self):
        """Testing Base64Field.to_python with None"""
        obj = Base64TestModel()
        value = obj._meta.get_field('field').to_python(None)

        self.assertIsNone(value)

    def test_value_to_string_with_base64_decoded_value(self):
        """Testing Base64Field.value_to_string with Base64DecodedValue"""
        obj = Base64TestModel(field=b'This is a test')
        value = obj._meta.get_field('field').value_to_string(obj)

        self.assertIs(type(value), str)
        self.assertEqual(value, 'VGhpcyBpcyBhIHRlc3Q=\n')

    def test_value_to_string_with_none(self):
        """Testing Base64Field.value_to_string with None"""
        obj = Base64TestModel(field=None)
        value = obj._meta.get_field('field').value_to_string(obj)

        self.assertIsNone(value)



"""Specialized descriptors/properties for classes."""

import warnings


class BaseProperty(object):
    """Base class for a custom property for a class.

    This is an optional base class that provides handy utilities that
    properties may need. For instance, being able to determine the name of
    the property's attribute on a class.
    """

    def get_attr_name(self, instance):
        """Return the name of this property's attribute.

        The value will be computed only once per property instance.

        Args:
            instance (object):
                The instance owning this property.

        Returns:
            str:
            The name of this property on the instance.
        """
        if not hasattr(self, '_attr_name'):
            cls = type(instance)
            self._attr_name = get_descriptor_attr_name(self, cls)

            assert self._attr_name is not None, (
                'Could not find the attribute for %r on %r' % (self, cls))

        return self._attr_name


class AliasProperty(BaseProperty):
    """A property that aliases to another property or attribute.

    Alias properties are used to automatically retrieve from another property
    on access, or to set a value on another property. It's useful when
    wanting to rename an attribute but continue to provide a deprecated
    version, or when creating an object that provides a set of compatibility
    attributes for use with legacy code.

    Alias properties can optionally emit a deprecation warning on use, in order
    to help in the process of migrating legacy code.
    """

    def __init__(self, prop_name, convert_to_func=None, convert_from_func=None,
                 deprecated=False, deprecation_warning=DeprecationWarning):
        """Initialize the property.

        Args:
            prop_name (str):
                The name of the property or attribute to read from and write
                it.

            convert_to_func (callable, optional):
                An optional function to call on a value before setting it on
                the aliased property name. This must take in the value as a
                parameter and return a value to set.

            convert_from_func (callable, optional):
                An optional function to call on a value after accessing it on
                the aliased property name and before returning to the caller.
                This must take in the value from the aliased property and
                return a value to return to the caller.

            deprecated (bool, optional):
                Whether to emit a deprecation warning when setting or accessing
                the property.

            deprecation_warning (type):
                The type of class to use for the deprecation warning. This
                should be a subclass of :py:class:`DeprecationWarning`.
        """
        self.prop_name = prop_name
        self.deprecated = deprecated
        self.deprecation_warning = deprecation_warning

        self._convert_to_func = convert_to_func
        self._convert_from_func = convert_from_func

    def __set__(self, instance, value):
        """Set a value on the property.

        This will convert the value (if ``convert_to_func`` was provided
        to this property) and set it on the aliased property.

        If this is a deprecated property, this will emit a warning.

        Args:
            instance (object):
                The instance owning this property.

            value (object):
                The value being set.
        """
        if self.deprecated:
            cls_name = type(instance).__name__
            warnings.warn('%s.%s is deprecated. Please set %s.%s instead.'
                          % (cls_name, self.get_attr_name(instance),
                             cls_name, self.prop_name),
                          self.deprecation_warning,
                          stacklevel=2)

        if value is not None and self._convert_to_func is not None:
            value = self._convert_to_func(value)

        setattr(instance, self.prop_name, value)

    def __get__(self, instance, owner):
        """Return the value of the property.

        This will retrieve the value from the aliased property, converting
        it (if ``convert_from_func`` was provided to this property), and
        return it.

        If this is a deprecated property, this will emit a warning.

        Args:
            instance (object):
                The instance owning this property.

            owner (type):
                The instance's class.

        Returns:
            object:
            The property value.
        """
        if self.deprecated:
            cls_name = type(instance).__name__
            warnings.warn('%s.%s is deprecated. Please access %s.%s instead.'
                          % (cls_name, self.get_attr_name(instance),
                             cls_name, self.prop_name),
                          self.deprecation_warning,
                          stacklevel=2)

        value = getattr(instance, self.prop_name)

        if value is not None and self._convert_from_func is not None:
            value = self._convert_from_func(value)

        return value


class TypedProperty(BaseProperty):
    """A property that enforces type safety.

    This property will ensure that only values that are compatible with a
    given type can be set. This ensures type safety and helps catch errors
    early.
    """

    def __init__(self, valid_types, default=None, allow_none=True):
        """Initialize the property.

        Args:
            valid_types (list of type):
                The types of values that are permitted to be set.

            default (object, optional):
                The default value, if one is not set.

            allow_none (bool, optional):
                Whether ``None`` values are allowed to be set.
        """
        self.valid_types = valid_types
        self.default = default
        self.allow_none = allow_none

    def __set__(self, instance, value):
        """Set a value on the property.

        This will check if the value is of a valid type, and then set it on
        the instance.

        Args:
            instance (object):
                The instance owning this property.

            value (object):
                The value being set.

        Raises:
            TypeError:
                The value is not of a valid type.
        """
        if value is None and not self.allow_none:
            raise TypeError('None is not a valid value for this property.')

        if value is not None and not isinstance(value, self.valid_types):
            raise TypeError('%s (%r) is not a valid type for this property.'
                            % (type(value).__name__, value))

        attr_name = self.get_attr_name(instance)
        instance.__dict__['_%s_typed' % attr_name] = value

    def __get__(self, instance, owner):
        """Return the value of the property.

        Args:
            instance (object):
                The instance owning this property.

            owner (type):
                The instance's class.

        Returns:
            object:
            The property value.
        """
        attr_name = self.get_attr_name(instance)
        return instance.__dict__.get('_%s_typed' % attr_name, self.default)


def get_descriptor_attr_name(descriptor, cls):
    """Return the name of a property/descriptor instance on a class.

    This will go through the class and all parent classes, looking for the
    property, and returning its attribute name. This is primarily intended
    to help with providing better error messages.

    Args:
        descriptor (object):
            The instance of the property/descriptor. For a proper value to
            be returned, this must exist on ``cls``.

        cls (type):
            The class owning the property.

    Returns:
        str:
        The name of the property/descriptor.
    """
    for attr_name, attr_value in cls.__dict__.items():
        if attr_value is descriptor:
            return attr_name

    for parent_cls in cls.__mro__:
        attr_name = get_descriptor_attr_name(descriptor, parent_cls)

        if attr_name is not None:
            return attr_name

    return None



"""Views for extensions (used in the admin UI)."""

from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect


@csrf_protect
@staff_member_required
def extension_list(request, extension_manager,
                   template_name='extensions/admin/extension_list.html'):
    if request.method == 'POST':
        if 'full-reload' in request.POST:
            extension_manager.load(full_reload=True)

        return HttpResponseRedirect('.')
    else:
        # Refresh the extension list.
        extension_manager.load()

        return render(request, template_name)


@csrf_protect
@staff_member_required
def configure_extension(request, ext_class, form_class, extension_manager,
                        template_name='extensions/configure_extension.html'):
    extension = extension_manager.get_enabled_extension(ext_class.id)

    if not extension or not extension.is_configurable:
        raise Http404

    if request.method == 'POST':
        form = form_class(extension, request.POST, request.FILES)

        if form.is_valid():
            form.save()

            return HttpResponseRedirect(request.path + '?saved=1')
    else:
        form = form_class(extension)

    return render(request, template_name, {
        'extension': extension,
        'form': form,
        'saved': request.GET.get('saved', 0),
    })



"""Unit tests for djblets.extensions.manager.SettingListWrapper."""

from django.conf import settings

from djblets.extensions.manager import SettingListWrapper
from djblets.testing.testcases import TestCase


class SettingListWrapperTests(TestCase):
    """Unit tests for djblets.extensions.manager.SettingListWrapper."""

    def test_loading_from_setting(self):
        """Testing SettingListWrapper constructor loading from settings"""
        settings.TEST_SETTING_LIST = ['item1', 'item2']
        wrapper = SettingListWrapper('TEST_SETTING_LIST', 'test setting list')

        self.assertEqual(wrapper.ref_counts.get('item1'), 1)
        self.assertEqual(wrapper.ref_counts.get('item2'), 1)

    def test_add_with_new_item(self):
        """Testing SettingListWrapper.add with new item"""
        settings.TEST_SETTING_LIST = []
        wrapper = SettingListWrapper('TEST_SETTING_LIST', 'test setting list')
        wrapper.add('item1')

        self.assertEqual(settings.TEST_SETTING_LIST, ['item1'])
        self.assertEqual(wrapper.ref_counts.get('item1'), 1)

    def test_add_with_existing_item(self):
        """Testing SettingListWrapper.add with existing item"""
        settings.TEST_SETTING_LIST = ['item1']
        wrapper = SettingListWrapper('TEST_SETTING_LIST', 'test setting list')
        wrapper.add('item1')

        self.assertEqual(settings.TEST_SETTING_LIST, ['item1'])
        self.assertEqual(wrapper.ref_counts.get('item1'), 2)

    def test_remove_with_ref_count_1(self):
        """Testing SettingListWrapper.remove with ref_count == 1"""
        settings.TEST_SETTING_LIST = ['item1']
        wrapper = SettingListWrapper('TEST_SETTING_LIST', 'test setting list')

        self.assertEqual(wrapper.ref_counts.get('item1'), 1)
        wrapper.remove('item1')

        self.assertEqual(settings.TEST_SETTING_LIST, [])
        self.assertFalse('item1' in wrapper.ref_counts)

    def test_remove_with_ref_count_gt_1(self):
        """Testing SettingListWrapper.remove with ref_count > 1"""
        settings.TEST_SETTING_LIST = ['item1']
        wrapper = SettingListWrapper('TEST_SETTING_LIST', 'test setting list')
        wrapper.add('item1')

        self.assertEqual(wrapper.ref_counts.get('item1'), 2)
        wrapper.remove('item1')

        self.assertEqual(settings.TEST_SETTING_LIST, ['item1'])
        self.assertEqual(wrapper.ref_counts.get('item1'), 1)



from django.db.models.query_utils import Q

from djblets.db.query import prefix_q
from djblets.testing.testcases import TestCase


class PrefixTests(TestCase):
    """Tests for djblets.db.query.prefix_q."""

    def test_simple(self):
        """Testing prefix_q prefixes simple expressions"""
        self.assertEqual(
            str(prefix_q('fk', Q(hello='goodbye'))),
            str(Q(fk__hello='goodbye')))

    def test_nested(self):
        """Testing prefix_q prefixes nested expressions"""
        self.assertEqual(
            str(prefix_q('fk',
                         Q(foo='foo') |
                         (Q(bar='bar') &
                          Q(baz='baz')))),
            str(Q(fk__foo='foo') |
                (Q(fk__bar='bar') &
                 Q(fk__baz='baz'))))

    def test_bytestring_result(self):
        """Testing that prefix_q generates byte strings for key names"""
        q = prefix_q('fk', Q(foo='bar'))
        self.assertEqual(len(q.children), 1)
        self.assertIs(type(q.children[0]), tuple)
        self.assertIsInstance(q.children[0][0], str)



"""Mixins for test cases that need to test enabled extensions."""

import pkg_resources
import uuid
import weakref
from contextlib import contextmanager

from djblets.extensions.extension import ExtensionInfo
from djblets.extensions.manager import (ExtensionManager,
                                        get_extension_managers,
                                        _extension_managers)
from djblets.extensions.models import RegisteredExtension


class _FakeProvider(pkg_resources.DefaultProvider):
    """A fake provider for a distribution.

    Version Added:
        2.3
    """

    egg_info = '/fake/path'

    def __init__(self, author='Example Author',
                 author_email='author@example.com',
                 description='Test description\u2049',
                 home_page='http://example.com',
                 project_name='ExampleProject', license_name='Drivers',
                 summary='Test summary', version='1.0'):
        """Initialize the FakeDistribution.

        Args:
            author (unicode):
                The package author.

            author_email (unicode):
                The package author's e-mail address.

            description (unicode):
                The package description.

            home_page (unicode):
                The package's URL.

            project_name (unicode):
                The package's name.

            license_name (unicode):
                The name of the package license.

            summary (unicode):
                The package summary.
        """
        self.metadata = {
            'Author': author,
            'Author-email': author_email,
            'Description': description,
            'Home-page': home_page,
            'Name': project_name,
            'License': license_name,
            'Summary': summary,
            'Version': version,
        }

    def _get(self, path):
        """Return the metadata content.

        This is the method that package resource providers must override to
        return metadata content for the package. It's expected to return
        byte strings, which will then be handled through the normal metadata
        functions.

        Returns:
            bytes:
            The package metadata content.
        """
        return ''.join(
            '%s: %s\n' % (field_name, value)
            for field_name, value in self.metadata.items()
        ).encode('utf-8')


class _FakeEntryPoint(object):
    """A fake entry point.

    This is modelled after :py:class:`pkg_resources.EntryPoint`.

    Version Added:
        2.3
    """

    def __init__(self, value, project_name, version='1.0', **metadata_kwargs):
        """Initialize the FakeEntryPoint.

        Args:
            value (object):
                The value to be returned when the entry point is loaded.

            project_name (unicode):
                The project name. This will be set in the metadata and as
                the distribution's name.

            **metadata_kwargs (dict):
                Keyword arguments to pass to the associated
                :py:class:`FakeProvider` constructor.
        """
        self._value = value
        self.dist = pkg_resources.Distribution(
            project_name=project_name,
            version=version,
            metadata=_FakeProvider(project_name=project_name,
                                   version=version,
                                   **metadata_kwargs))

    def load(self):
        """Load the entry point.

        Returns:
            object: The value specified at initialization time.
        """
        return self._value


class DefaultTestsExtensionManager(ExtensionManager):
    """Default extension manager for unit tests.

    Version Added:
        2.3
    """


class ExtensionTestCaseMixin(object):
    """Unit tests mixin for more easily testing extensions.

    This will do the hard work of creating the fake registration information
    needed for an extension and to instantiate an instance for testing.

    Subclasses need to define :py:attr:`extension_class` and may want to
    implement :py:meth:`get_extension_manager` (by default, the first
    registered extension manager will be used).

    Projects may want to provide their own subclass for their extensions to use
    that implements :py:meth:`get_extension_manager`, so extensions won't have
    to.

    Version Changed:
        2.3:
        Extensions and extension managers are now cleaned up and removed after
        test runs.

        Subclasses can more easily set up default extension state, or create
        new extensions as needed.

    Attributes:
        extension_mgr (djblets.extensions.manager.ExtensionManager):
            The extension manager owning the extension. Tests can use this to
            manually enable/disable the extension, if needed.

        extension (djblets.extensions.extension.Extension):
            The extension instance being tested.
    """

    #: The extension class to test.
    extension_class = None

    #: Optional metadata to use for the extension information.
    extension_metadata = {
        'Version': '1.0',
    }

    #: Optional package name to use for the extension information.
    extension_package_name = 'TestPackage'

    #: Whether to enable the extension by default.
    #:
    #: Version Added:
    #:     2.3
    #:
    #: Type:
    #:     bool
    extension_enable_by_default = True

    #: The default extension manager to use if one is not instantiated.
    #:
    #: This will be created only if the consuming project hasn't already
    #: instantiated an extension manager.
    #:
    #: Version Added:
    #:     2.3
    #:
    #: Type:
    #:     type
    default_extension_manager_cls = DefaultTestsExtensionManager

    #: Whether to allow use of an existing registered extension manager.
    #:
    #: If disabled, a new extension manager will be created for the test run.
    #: This is enabled by default for backwards-compatibility.
    #:
    #: Version Added:
    #:     2.3
    #:
    #: Type:
    #:     bool
    allow_existing_extension_manager = True

    def setUp(self):
        super(ExtensionTestCaseMixin, self).setUp()

        self.extension_mgr = self.get_extension_manager()

        self._extension_classes = []
        self._extension_instances = []
        self._extensions = []
        self._orig_extension_managers = set(get_extension_managers())

        if self.extension_class is not None:
            self.extension = self.setup_extension(
                extension_cls=self.extension_class,
                package_name=self.extension_package_name,
                metadata=self.extension_metadata,
                enable=self.extension_enable_by_default)
        else:
            self.extension = None

    def tearDown(self):
        super(ExtensionTestCaseMixin, self).tearDown()

        self.reset_extensions()

    def reset_extensions(self):
        """Unregister and clean up all extensions and extension managers.

        This will disable and unregister any extensions that have been set
        up using :py:meth:`setup_extension`. It will also shut down and
        unregister all extension managers created during the test run.

        This is normally called when a unit test has finished. Tests can
        explicitly call this to reset the extension management state.

        Version Added:
            2.3
        """
        for extension_ref in self._extension_instances:
            extension = extension_ref()

            if extension is not None and getattr(extension, 'info', None):
                extension_mgr = extension.extension_manager

                if extension.info.enabled:
                    # Manually shut down the extension first, before we have
                    # the extension manager disable it. This will help ensure
                    # we have the right state up-front.
                    extension.shutdown()

                extension_mgr.disable_extension(extension.id)

        # If the unit test created any new extension managers, remove them now.
        new_extension_mgrs = (
            (set(get_extension_managers()) - self._orig_extension_managers) |
            {self.extension_mgr})

        for extension_mgr in new_extension_mgrs:
            extension_mgr.clear_sync_cache()
            extension_mgr.shutdown()

            for extension_cls_ref in self._extension_classes:
                extension_cls = extension_cls_ref()

                if (extension_cls is not None and
                    extension_cls.id in extension_mgr._extension_classes):
                    del extension_mgr._extension_classes.pop[extension_cls.id]
                    del self._extension_classes[extension_cls.id]
                    break

            _extension_managers.pop(id(extension_mgr), None)

        self._extension_classes = []
        self._extension_instances = []
        self._extensions = []
        self._orig_extension_managers = set(get_extension_managers())
        self.extension_mgr = None
        self.extension = None

    @contextmanager
    def scanned_extensions(self, extension_classes=[], extension_mgr=None):
        """Make the specified extension classes available for scanning.

        This should be called before calling :py:meth:`ExtensionManager.load()
        <djblets.extensions.manager.ExtensionManager.load>` in order to ensure
        the specified list of extension classes are available to be scanned.

        Version Added:
            2.3

        Args:
            extension_classes (list, optional):
                The list of extensions to be made available.

                Each entry can be an extension class, or a tuple of
                ``(extension_class, package_name)``.

                If any extension is not registered, it will be assigned
                a default package name of :py:attr:`extension_package_name`.

            extension_mgr (djblets.extensions.manager.ExtensionManager,
                           optional):
                An explicit extension manager that will be used to scan,
                rather than the test suite's :py:attr:`extension_mgr`
                attribute.

        Context:
            The provided extensions will be available for scanning.
        """
        assert hasattr(self, 'spy_on'), (
            '%r must inherit from kgb.SpyAgency.' % type(self))

        if extension_mgr is None:
            extension_mgr = self.extension_mgr

        assert extension_mgr is not None

        def _entrypoint_iterator(_self):
            results = []

            for entry in extension_classes:
                if isinstance(entry, tuple):
                    extension_cls, package_name = entry
                else:
                    extension_cls = entry
                    info = getattr(extension_cls, 'info', None)

                    if info is not None:
                        package_name = info.package_name
                    else:
                        package_name = self.extension_package_name

                results.append(_FakeEntryPoint(extension_cls,
                                               project_name=package_name))

            return results

        spy = self.spy_on(extension_mgr._entrypoint_iterator,
                          call_fake=_entrypoint_iterator)

        try:
            yield
        finally:
            spy.unspy()

    def setup_extension(self,
                        extension_cls=None,
                        package_name=None,
                        metadata=None,
                        enable=True,
                        extension_mgr=None):
        """Set up a new extension instance.

        This will construct an new extension of the provided type with the
        provided information.

        The extension will be registered in the extension manager, and cleaned
        up when the unit test has finished.

        Version Added:
            2.3

        Args:
            extension_cls (type, optional):
                The extension's class.

                Defaults to :py:attr:`extension_class`.

            package_name (unicode, optional):
                The package name assigned as the owner of the extension.

                Defaults to :py:attr:`extension_package_name`.

            metadata (dict, optional):
                The extension package metadata.

                Defaults to :py:attr:`metadata`.

            enable (bool, optional):
                Whether to enable the extension.

                If not enabled, an instance will still be returned, but it
                won't be registered as enabled in the extension manager.

            extension_mgr (djblets.extensions.manager.ExtensionManager,
                           optional):
                An explicit extension manager to register this extension in.

                Defaults to the result of :py:meth:`get_extension_manager`.

        Returns:
            djblets.extensions.extension.Extension:
            The extension instance, if enabling. Otherwise, ``None`` will be
            returned.
        """
        if extension_cls is None:
            extension_cls = self.extension_class

        if package_name is None:
            package_name = self.extension_package_name

        if metadata is None:
            metadata = self.extension_metadata

        if extension_cls is None:
            raise ValueError('extension_cls cannot be None.')

        if package_name is None:
            raise ValueError('package_name cannot be None.')

        if metadata is None:
            raise ValueError('metadata cannot be None.')

        if extension_cls in self._extension_classes:
            raise ValueError(
                '%r has already been registered during this test run.'
                % extension_cls)

        extension_mgr = extension_mgr or self.extension_mgr

        # We want to override all the information, even if a previous test
        # already set it. The metadata may be different, and the registration
        # definitely needs to be replaced (as it contains extension settings).
        extension_id = '%s.%s' % (extension_cls.__module__,
                                  extension_cls.__name__)

        extension_cls.id = extension_id
        extension_cls.info = ExtensionInfo(ext_class=extension_cls,
                                           package_name=package_name,
                                           metadata=metadata)

        extension_cls.registration = RegisteredExtension.objects.create(
            class_name=extension_id,
            name=extension_cls.info.name,
            enabled=True,
            installed=True)

        # We're going to manually inject the extension, instead of calling
        # load(), since it might not be found otherwise.
        extension_mgr._extension_classes[extension_id] = extension_cls
        self._extension_classes.append(weakref.ref(extension_cls))

        if enable:
            extension = extension_mgr.enable_extension(extension_id)
            assert extension is not None

            self._extension_instances.append(weakref.ref(extension))
        else:
            extension = None

        return extension

    def get_extension_manager(self):
        """Return the extension manager used for the tests.

        Subclasses may want to override this to pick a specific extension
        manager, if the project uses more than one. The default behavior is
        to return the first registered extension manager, or create one if
        one is not available.

        If :py:attr:`self.allow_existing_extension_manager` is ``False``, this
        will always create a new extension manager.

        Version Changed:
            2.3:
            This is now guaranteed to return an extension manager instance,
            and that instance is guaranteed to be cleaned up after the test
            has finished.

        Returns:
            djblets.extensions.manager.ExtensionManager:
            The extension manager used for tests.
        """
        extension_manager = None

        if self.allow_existing_extension_manager:
            try:
                extension_manager = get_extension_managers()[0]
            except IndexError:
                pass

        if extension_manager is None:
            extension_manager = self.default_extension_manager_cls(
                key='tests-%s' % uuid.uuid4())

        return extension_manager



from django.apps.registry import apps
from django.conf import settings
from django.template import Context, Template, TemplateSyntaxError, engines
from django.template.loader import get_template
from django.template.loaders.cached import Loader as CachedLoader

from djblets.template.caches import (clear_template_caches,
                                     clear_template_tag_caches)
from djblets.testing.testcases import TestCase


class CachesTests(TestCase):
    """Unit tests for djblets.template.caches."""

    def test_clear_template_tag_caches(self):
        """Testing clear_template_tag_caches"""

        def _check_state(enabled):
            if enabled:
                self.assertEqual(
                    Template(template_str).render(Context({})),
                    'Hello, world!')
            else:
                with self.assertRaisesRegex(TemplateSyntaxError,
                                            'is not a (valid|registered) tag '
                                            'library'):
                    Template(template_str).render(Context({}))

        template_str = (
            '{% load template_tests %}'
            '{% my_test_template_tag %}'
        )

        # Sanity-check that the template tag module isn't registered.
        _check_state(enabled=False)

        # Enable a new template tag library.
        old_installed_apps = settings.INSTALLED_APPS

        settings.INSTALLED_APPS = list(old_installed_apps) + [
            'djblets.template.tests',
        ]

        apps.set_installed_apps(settings.INSTALLED_APPS)

        try:
            clear_template_tag_caches()
            _check_state(enabled=True)
        finally:
            settings.INSTALLED_APPS = old_installed_apps
            apps.unset_installed_apps()

        clear_template_tag_caches()
        _check_state(enabled=False)

        # Other libraries should still work.
        Template('{% load djblets_js djblets_extensions %}').render(
            Context({}))

    def test_clear_template_caches(self):
        """Testing clear_template_caches"""
        # Load a template to populate the cache.
        get_template('avatars/avatar.html')

        template_loader = engines.all()[0].engine.template_loaders[0]

        self.assertTrue(isinstance(template_loader, CachedLoader))
        self.assertNotEqual(template_loader.get_template_cache, {})

        clear_template_caches()

        self.assertEqual(template_loader.get_template_cache, {})



"""Compatibility functions for older Django cache backend configuration.

These functions are used to maintain compatibility or transition settings from
older versions of Django.
"""

import logging

from django.core.cache import DEFAULT_CACHE_ALIAS, InvalidCacheBackendError

try:
    from django.core.cache import parse_backend_uri
except ImportError:
    from djblets.util.compat.django.core.cache import parse_backend_uri


logger = logging.getLogger(__name__)


BACKEND_CLASSES = {
    'db': 'db.DatabaseCache',
    'dummy': 'dummy.DummyCache',
    'file': 'filebased.FileBasedCache',
    'locmem': 'locmem.LocMemCache',
    'memcached': 'memcached.PyMemcacheCache',
}

RENAMED_BACKENDS = {
    'django.core.cache.backends.memcached.CacheClass':
        'django.core.cache.backends.memcached.PyMemcacheCache',
    'django.core.cache.backends.memcached.MemcachedCache':
        'django.core.cache.backends.memcached.PyMemcacheCache',
}


def normalize_cache_backend(cache_backend, cache_name=DEFAULT_CACHE_ALIAS):
    """Return a new-style ``CACHES`` dictionary from any given cache_backend.

    Over time, Django has had support for two formats for a cache backend. The
    old-style :django:setting:`CACHE_BACKEND` string, and the new-style
    :django:setting:`CACHES` dictionary.

    This function will accept either as input and return a dictionary
    containing a single normalized entry for the given cache backend. This can
    be merged into the existing :django:setting:`CACHES` setting.

    If a :django:setting:`CACHES` dictionary is passed, the "default" cache
    will be the result.

    Args:
        cache_backend (dict or str):
            The new-style or old-style cache backend dictionary or str to
            normalize.

        cache_name (str):
            The name of the cache backend to look up in ``cache_backend``, if
            a new-style dictionary is provided.

    Returns:
        A new-style cache backend dictionary containing the single cache
        backend referenced. If there were any parsing issues, an empty
        dictionary will be returned.
    """
    if not cache_backend:
        return {}

    if isinstance(cache_backend, dict):
        backend_info = cache_backend.get(cache_name, {})
        backend_name = backend_info.get('BACKEND')

        if backend_name in RENAMED_BACKENDS:
            backend_info['BACKEND'] = RENAMED_BACKENDS[backend_name]

        return backend_info

    try:
        engine, host, params = parse_backend_uri(cache_backend)
    except InvalidCacheBackendError as e:
        logger.error('Invalid cache backend (%s) found while loading '
                     'siteconfig: %s',
                     cache_backend, e)
        return {}

    if engine in BACKEND_CLASSES:
        engine = 'django.core.cache.backends.%s' % BACKEND_CLASSES[engine]
    else:
        engine = '%s.CacheClass' % engine

    defaults = {
        'BACKEND': engine,
        'LOCATION': host,
    }
    defaults.update(params)

    return defaults



"""HTTP-related utilities."""

import hashlib
from urllib.parse import urlencode

from django.http import HttpResponse, HttpResponseNotModified
from django.utils.encoding import force_bytes

from djblets.util.dates import http_date


class HttpResponseNotAcceptable(HttpResponse):
    status_code = 406


def set_last_modified(response, timestamp):
    """
    Sets the Last-Modified header in a response based on a DateTimeField.
    """
    response['Last-Modified'] = http_date(timestamp)


def get_modified_since(request, last_modified):
    """
    Checks if a Last-Modified timestamp is newer than the requested
    HTTP_IF_MODIFIED_SINCE from the browser. This can be used to bail
    early if no updates have been performed since the last access to the
    page.

    This can take a DateField, datetime, HTTP date-formatted string, or
    a function for the last_modified timestamp. If a function is passed,
    it will only be called if the HTTP_IF_MODIFIED_SINCE header is present.
    """
    if_modified_since = request.META.get('HTTP_IF_MODIFIED_SINCE', None)

    if if_modified_since is not None:
        if callable(last_modified):
            last_modified = last_modified()

        return (if_modified_since == http_date(last_modified))

    return False


def set_etag(response, etag):
    """
    Sets the ETag header in a response.
    """
    response['ETag'] = etag


def encode_etag(etag):
    """Encode a string as a SHA1 value, for use in an ETag.

    Args:
        etag (bytes or unicode):
            The ETag to encode.

    Returns:
        bytes:
        The encoded ETag.
    """
    return hashlib.sha1(force_bytes(etag)).hexdigest()


def etag_if_none_match(request, etag):
    """
    Checks if an ETag matches the If-None-Match header sent by the browser.
    This can be used to bail early if no updates have been performed since
    the last access to the page.
    """
    return etag == request.META.get('HTTP_IF_NONE_MATCH', None)


def etag_if_match(request, etag):
    """
    Checks if an ETag matches the If-Match header sent by the browser. This
    is used by PUT requests to to indicate that the update should only happen
    if the specified ETag matches the header.
    """
    return etag == request.META.get('HTTP_IF_MATCH', None)


def build_not_modified_from_response(response):
    """Build a Not Modified response from an existing response.

    This will create a new :py:class:`~django.http.HttpResponseNotModified`
    containing headers from an original response. It should be used in
    conjunction with functions like :py:func:`etag_if_none_match` when
    checking against an otherwise valid response that's been built.

    Args:
        response (django.http.HttpResponse):
            The HTTP response to build the Not Modified response from.

    Returns:
        django.http.HttpResponseNotModified:
        The new Not Modified response.
    """
    new_response = HttpResponseNotModified()

    # Copy cache-related headers from the original response into the new one.
    for header_name in ('Cache-Control', 'Content-Location', 'Date', 'ETag',
                        'Expires', 'Vary'):
        try:
            new_response[header_name] = response[header_name]
        except KeyError:
            pass

    return new_response


def get_http_accept_lists(request):
    """Returns lists of mimetypes from the request's Accept header.

    This will return two lists, a list of acceptable mimetypes in order
    of requested priority, and a list of unacceptable mimetypes.
    """
    # Check cached copies for this in the request so we only ever do it once.
    if (hasattr(request, 'djblets_acceptable_mimetypes') and
        hasattr(request, 'djblets_unacceptable_mimetypes')):
        return (request.djblets_acceptable_mimetypes,
                request.djblets_unacceptable_mimetypes)

    acceptable_mimetypes = []
    unacceptable_mimetypes = []

    for accept_item in request.META.get('HTTP_ACCEPT', '').strip().split(','):
        parts = accept_item.strip().split(";")
        mimetype = parts[0]
        priority = 1.0

        for part in parts[1:]:
            try:
                key, value = part.split('=')
            except ValueError:
                # There's no '=' in that part.
                continue

            if key == 'q':
                try:
                    priority = float(value)
                except ValueError:
                    # The value isn't a number.
                    continue

        if priority == 0:
            unacceptable_mimetypes.append(mimetype)
        else:
            acceptable_mimetypes.append((mimetype, priority))

    acceptable_mimetypes.sort(key=lambda x: x[1], reverse=True)
    acceptable_mimetypes = [m[0] for m in acceptable_mimetypes]

    setattr(request, 'djblets_acceptable_mimetypes', acceptable_mimetypes)
    setattr(request, 'djblets_unacceptable_mimetypes', unacceptable_mimetypes)

    return acceptable_mimetypes, unacceptable_mimetypes


def get_http_requested_mimetype(request, supported_mimetypes):
    """Gets the mimetype that should be used for returning content.

    This is based on the client's requested list of mimetypes (in the
    HTTP Accept header) and the supported list of mimetypes that can be
    returned in this request.

    If a valid mimetype that can be used is found, it will be returned.
    Otherwise, None is returned, and the caller is expected to return
    HttpResponseNotAccepted.
    """
    acceptable_mimetypes, unacceptable_mimetypes = \
        get_http_accept_lists(request)

    supported_mimetypes_set = set(supported_mimetypes)
    acceptable_mimetypes_set = set(acceptable_mimetypes)
    unacceptable_mimetypes_set = set(unacceptable_mimetypes)

    if not supported_mimetypes_set.intersection(acceptable_mimetypes_set):
        # None of the requested mimetypes are in the supported list.
        # See if there are any mimetypes that are explicitly forbidden.
        if '*/*' in unacceptable_mimetypes:
            acceptable_mimetypes = []
            unacceptable_mimetypes = supported_mimetypes
        else:
            acceptable_mimetypes = [
                mimetype
                for mimetype in supported_mimetypes
                if mimetype not in unacceptable_mimetypes_set
            ]

    if acceptable_mimetypes:
        for mimetype in acceptable_mimetypes:
            if mimetype in supported_mimetypes:
                return mimetype

    # We didn't find any mimetypes that are on the supported list.
    # We need to choose a default now.
    for mimetype in supported_mimetypes:
        if mimetype not in unacceptable_mimetypes:
            return mimetype

    return None


def is_mimetype_a(mimetype, parent_mimetype):
    """Returns whether or not a given mimetype is a subset of another.

    This is generally used to determine if vendor-specific mimetypes is
    a subset of another type. For example,
    :mimetype:`application/vnd.djblets.foo+json` is a subset of
    :mimetype:`application/json`.
    """
    parts = mimetype.split('/')
    parent_parts = parent_mimetype.split('/')

    return (parts[0] == parent_parts[0] and
            (parts[1] == parent_parts[1] or
             parts[1].endswith('+' + parent_parts[1])))


def get_url_params_except(query, *params):
    """Return a URL query string that filters out some params.

    This is used often when one wants to preserve some GET parameters and not
    others.
    """
    return urlencode([
        (key.encode('utf-8'), value.encode('utf-8'))
        for key, value in query.items()
        if key not in params
    ])



from django.db import models


class MyTestExtensionModel(models.Model):
    test_field = models.CharField(max_length=16)

    class Meta(object):
        app_label = 'model_tests'
        db_table = 'model_tests_testextensionmodel'



"""OAuth2 scope generation for WebAPI resources."""

import logging
import threading
from collections import defaultdict
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext as _

from djblets.extensions.manager import get_extension_managers
from djblets.extensions.signals import (extension_enabled,
                                        extension_disabled)
from djblets.webapi.resources.mixins.oauth2_tokens import (
    ResourceOAuth2TokenMixin)


logger = logging.getLogger(__name__)

_enable_lock = threading.Lock()
_scopes = None


def get_scope_dictionary():
    """Return the scope dictionary.

    This method requires :setting:`WEB_API_ROOT_RESOURCE` setting to point to
    an instance of the root resource for your WebAPI. Optionally, the
    :setting:`WEB_API_SCOPE_DICT_CLASS` (defaulting to
    :py:class:`djblets.webapi.oauth2_scopes.WebAPIScopeDictionary`) can be set
    to specialize scope generation.

    If the :py:mod:`djblets.extensions` app is being used, then the
    :setting:`WEB_API_SCOPE_DICT_CLASS` setting should be set to
    :py:class:`djblets.webapi.oauth2_scopes.ExtensionEnabledWebAPIScopeDictionary`.

    Returns:
        WebAPIScopeDictionary:
        The scope dictionary.
    """
    global _scopes

    if _scopes is None:
        scopes_cls_name = getattr(
            settings, 'WEB_API_SCOPE_DICT_CLASS',
            'djblets.webapi.oauth2_scopes.WebAPIScopeDictionary')

        scopes_cls_mod, scopes_cls_attr = scopes_cls_name.rsplit('.', 1)

        try:
            scopes_cls = getattr(import_module(scopes_cls_mod),
                                 scopes_cls_attr)
        except (AttributeError, ImportError) as e:
            raise ImproperlyConfigured(
                'settings.WEB_API_SCOPE_DICT_CLASS %r could not be imported: '
                '%s'
                % (scopes_cls_name, e)
            )

        try:
            root_resource_path = getattr(settings, 'WEB_API_ROOT_RESOURCE')
        except AttributeError:
            raise ImproperlyConfigured(
                'settings.WEB_API_ROOT_RESOURCE is required.'
            )

        root_resource_mod, root_resource_attr = root_resource_path.rsplit('.',
                                                                          1)

        try:
            root_resource = getattr(import_module(root_resource_mod),
                                    root_resource_attr)
        except (AttributeError, ImportError) as e:
            raise ImproperlyConfigured(
                'settings.WEB_API_ROOT_RESOURCE %r could not be imported: %s'
                % (root_resource_path, e)
            )

        _scopes = scopes_cls(root_resource)

    return _scopes


class WebAPIScopeDictionary(object):
    """A Web API scope dictionary.

    This class knows how to build a list of available scopes from the WebAPI
    resource tree at runtime.

    By default, it will only have to walk the API tree once, after which the
    value can be cached.
    """

    def __init__(self, root_resource):
        """Initialize the scope dictionary.

        Args:
            root_resource (djblets.webapi.resources.base.WebAPIResource):
                The root resource to walk in order to build scopes.
        """
        self.resource_trees = {root_resource}
        self._update_lock = threading.Lock()
        self._scope_dict = {}

    @property
    def scope_dict(self):
        """The dictionary of scopes defined by this dictionary.

        The value is cached so that it will only be recomputed when the
        dictionary is updated.
        """
        if not self._scope_dict:
            with self._update_lock:
                if not self._scope_dict:
                    self._walk_resources(self.resource_trees)

                    assert self._scope_dict

        return self._scope_dict

    def keys(self):
        """Iterate through all keys in the dictionary.

        This is used by oauth2_provider when on Python 3.x to get the list
        of scope keys.

        Yields:
            unicode:
            The key for each scope.
        """
        return self.scope_dict.keys()

    def clear(self):
        """Clear all scopes from the dictionary.

        The next attempt at fetching scopes will repopulate the dictionary
        from scratch.
        """
        self._scope_dict.clear()

    def _walk_resources(self, resources):
        """Traverse the given resource trees and add the appropriate scopes.

        Args:
            resources (list of djblets.webapi.resources.base.WebAPIResource):
                The resources to generate scopes for. The children of these
                resources will also be traversed.

        Returns:
            list of unicode:
            The list of scopes added by walking the given resources.
        """
        for resource in resources:
            self._walk_resources(resource.list_child_resources)
            self._walk_resources(resource.item_child_resources)

            scope_to_methods = defaultdict(list)

            if not isinstance(resource, ResourceOAuth2TokenMixin):
                logging.warning(
                    'Resource %r does not inherit from '
                    'djblets.webapi.resources.mixins.oauth2_tokens.'
                    'ResourceOAuth2TokenMixin: it will not be accessible with '
                    'OAuth2 tokens. It is recommended that all your resources '
                    'inherit from a base class that includes this mixin.',
                    type(resource),
                )
                continue

            if not resource.oauth2_token_access_allowed:
                continue

            for method in resource.allowed_methods:
                try:
                    suffix = resource.HTTP_SCOPE_METHOD_MAP[method]
                except KeyError:
                    logger.error('Unknown HTTP method %s not present in '
                                 'HTTP_SCOPE_METHOD_MAP: a scope will not be '
                                 'generated for this method.',
                                 method)
                    continue

                scope_to_methods[suffix].append(method)

            for suffix, methods in scope_to_methods.items():
                scope_name = '%s:%s' % (resource.scope_name, suffix)

                self._scope_dict[scope_name] = (
                    _('Ability to perform HTTP %(methods)s on the %(name)s '
                      'resource')
                    % {
                        'methods': ', '.join(sorted(methods)),
                        'name': resource.name,
                    }
                )

    def __getitem__(self, key):
        """Return the description of a given scope.

        Args:
            key (unicode):
                The scope's key.

        Returns:
            unicode:
            The scope's description.

        Raises:
            KeyError:
                The scope's key was not in the dictionary.
        """
        return self.scope_dict[key]

    def __contains__(self, key):
        """Return whether the dictionary has a particular scope.

        Args:
            key (unicode):
                The scope's key.

        Returns:
            bool:
            ``True`` if the scope is in the dictionary.
        """
        return key in self.scope_dict

    def __repr__(self):
        """Return a string representation of this object.

        Returns:
            unicode:
            A string representation of this object.
        """
        return '%s(%r)' % (type(self).__name__,
                           list(self.scope_dict.keys()))


class ExtensionEnabledWebAPIScopeDictionary(WebAPIScopeDictionary):
    """A Web API scopes dictionary that supports extensions.

    This scope dictionary is only required if your app is using the
    :py:mod:`djblets.extensions` app.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the scope dictionary.

        This adds signal handlers to ensure the dictionary stays up to date
        when extensions are initialized and uninitialized.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the the parent class.
        """
        super(ExtensionEnabledWebAPIScopeDictionary, self).__init__(*args,
                                                                    **kwargs)

        extension_enabled.connect(self._on_extension_enabled)
        extension_disabled.connect(self._on_extension_disabled)

        for manager in get_extension_managers():
            for extension in manager.get_enabled_extensions():
                self._on_extension_enabled(extension=extension)

    def _on_extension_enabled(self, extension, **kwargs):
        """Traverse an extensions resource trees when it is enabled.

        Args:
            extension (djblets.extensions.extension.Extension):
                The extension class that was enabled.

            **kwargs (dict):
                Ignored keyword arguments from the signal.
        """
        if extension.resources:
            self.resource_trees.update(extension.resources)
            self.clear()

    def _on_extension_disabled(self, extension, **kwargs):
        """Remove an extensions scopes when it is uninitialized.

        Args:
            ext_class (type):
                The extension class that was uninitialized.

            **kwargs (dict):
                Ignored keyword arguments from the signal.
        """
        if extension.resources:
            self.resource_trees.difference_update(extension.resources)
            self.clear()


def enable_web_api_scopes(*args, **kwargs):
    """Enable WebAPI scopes.

    The ``oauth2_provider.settings.oauth2_settings`` object will be patched so
    that scopes are correctly cached. The cache will be lazily updated after
    the scopes are updated.

    Args:
        *args (tuple):
            Ignored positional arguments.

        **kwargs (dict):
            Ignored keyword arguments.
    """
    with _enable_lock:
        if not isinstance(settings.OAUTH2_PROVIDER['SCOPES'],
                          WebAPIScopeDictionary):
            settings.OAUTH2_PROVIDER['SCOPES'] = get_scope_dictionary()



"""Settings managers for avatar service registries."""


class AvatarSettingsManager(object):
    """The settings manager is responsible for loading and saving settings.

    Each user can have different avatar configuration and the settings
    manager is responsible for loading and saving per-user configuration for
    services.

    This class is intended to be sub-classed to be able to provide
    configuration management for individual users.
    """

    def __init__(self, user):
        """Initialize the settings manager.

        Args:
            user (django.contrib.auth.models.User):
                The user.
        """
        self.user = user

    @property
    def avatar_service_id(self):
        """The service ID for the user's selected avatar service.

        Returns:
            unicode:
            The avatar service ID for the user's selected avatar service, or
            ``None`` if they have not selected one.
        """
        raise NotImplementedError('%s does not implement avatar_service_id'
                                  % type(self).__name__)

    @avatar_service_id.setter
    def avatar_service_id(self, avatar_service_id):
        """Set the avatar service ID for the user.

        Args:
            avatar_service_id (unicode):
                The ID of the :py:class:`avatar service
                <djblets.avatars.services.base.AvatarService>` to set.
        """
        raise NotImplementedError('%s does not implement avatar_service_id'
                                  % type(self).__name__)

    @property
    def configuration(self):
        """The user's configuration for the service.

        This must be implemented in a subclasses.

        Returns:
            dict:
            The user's configuration.
        """
        raise NotImplementedError('%s does not implement configuration'
                                  % type(self).__name__)

    @configuration.setter
    def configuration(self, settings):
        """Set the user's configuration for the service.

        This must be implemented in a subclass.

        Args:
            settings (dict):
                The settings to save.
        """
        raise NotImplementedError('%s does not implement configuration'
                                  % type(self).__name__)

    def configuration_for(self, avatar_service_id):
        """Get the configuration for the requested avatar service.

        Args:
            avatar_service_id (unicode):
                The ID of the :py:class:`avatar service
                <djblets.avatars.services.base.AvatarService>` to retrieve
                configuration for.
        """
        raise NotImplementedError('%s does not implement configuration_for'
                                  % type(self).__name__)

    def save(self):
        """Save the configuration.

        This must be implemented in a subclass.
        """
        raise NotImplementedError('%s does not implement save()')



"""Token generators for creating access tokens.

Version Added:
    3.0
"""

from djblets.registries.importer import lazy_import_registry


token_generator_registry = \
    lazy_import_registry('djblets.secrets.token_generators.registry',
                         'TokenGeneratorRegistry')
